# Copyright (c) 2025, Jay Shah, Ganesh Bikshandi, Ying Zhang, Vijay Thakkar, Pradeep Ramani, Tri Dao.
# SPDX-License-Identifier: MIT
# A reimplementation of https://github.com/Dao-AILab/flash-attention/blob/main/hopper/flash_bwd_postprocess_kernel.h
# from Cutlass C++ to Cute-DSL.
import math
from typing import Callable, Optional, Type

import cuda.bindings.driver as cuda

import cutlass
import cutlass.cute as cute
import cutlass.utils.blackwell_helpers as sm100_utils_basic
from cutlass.cute.nvgpu import cpasync
from cutlass import Float32, const_expr
from cutlass.utils import LayoutEnum

from cudnn.block_sparse_attention.csrc.utils import copy_utils, layout_utils

from cudnn.block_sparse_attention.csrc.utils import kernel_utils as utils
from cudnn.block_sparse_attention.csrc.utils.cute_dsl_utils import assume_tensor_aligned
from cudnn.block_sparse_attention.csrc.utils.seqlen_info import SeqlenInfoQK
import cutlass.cute.nvgpu.tcgen05 as tcgen05
from cudnn.block_sparse_attention.csrc.utils.cute_dsl_utils import ParamsBase
from cudnn.block_sparse_attention.csrc.utils.tile_scheduler import (
    SingleTileScheduler,
    SingleTileVarlenScheduler,
    TileSchedulerArguments,
)


class BlockSparseAttnBackwardPostprocess:
    def __init__(
        self,
        dtype: Type[cutlass.Numeric],
        head_dim: int,
        arch: int,
        tile_m: int = 128,
        dQ_swapAB: bool = False,
    ):
        """
        :param head_dim: head dimension
        :type head_dim: int
        :param tile_m: m block size
        :type tile_m: int
        """
        self.dtype = dtype
        self.tile_m = tile_m
        assert arch // 10 in [10, 11], "Only Blackwell (10.x, 11.x) is supported"
        # padding head_dim to a multiple of 32 as k_block_size
        hdim_multiple_of = 32
        self.tile_hdim = int(math.ceil(head_dim / hdim_multiple_of) * hdim_multiple_of)
        self.num_threads = 128
        self.dQ_swapAB = dQ_swapAB

    def _get_tiled_mma(self):
        tiled_mma = sm100_utils_basic.make_trivial_tiled_mma(
            self.dtype,
            self.dtype,
            cute.nvgpu.OperandMajorMode.MN,  # dS_major_mode
            cute.nvgpu.OperandMajorMode.MN,  # Kt_major_mode
            Float32,
            tcgen05.CtaGroup.ONE,
            (self.tile_m, self.tile_hdim),
        )
        return tiled_mma

    def _setup_attributes(self):
        # ///////////////////////////////////////////////////////////////////////////////
        # GMEM Tiled copy:
        # ///////////////////////////////////////////////////////////////////////////////
        # Thread layouts for copies
        universal_copy_bits = 128
        async_copy_elems_accum = universal_copy_bits // Float32.width
        atom_async_copy_accum = cute.make_copy_atom(
            cpasync.CopyG2SOp(cache_mode=cute.nvgpu.LoadCacheMode.GLOBAL),
            Float32,
            num_bits_per_copy=universal_copy_bits,
        )
        # We don't do bound checking for the gmem -> smem load so we just assert here.
        assert (self.tile_m * self.tile_hdim // async_copy_elems_accum) % self.num_threads == 0
        self.g2s_tiled_copy_dQaccum = cute.make_tiled_copy_tv(
            atom_async_copy_accum,
            cute.make_layout(self.num_threads),
            cute.make_layout(async_copy_elems_accum),
        )
        num_s2r_copy_elems = 4
        self.dQ_reduce_ncol = 32
        dQaccum_reduce_stage = self.tile_hdim // self.dQ_reduce_ncol
        self.s2r_tiled_copy_dQaccum = copy_utils.tiled_copy_1d(Float32, self.num_threads, num_s2r_copy_elems)
        self.sdQaccum_layout = cute.make_layout((self.tile_m * self.tile_hdim // dQaccum_reduce_stage, dQaccum_reduce_stage))

        num_copy_elems = 128 // self.dtype.width
        threads_per_row = math.gcd(128, self.tile_hdim) // num_copy_elems
        self.gmem_tiled_copy_dQ = copy_utils.tiled_copy_2d(self.dtype, threads_per_row, self.num_threads, num_copy_elems)
        # ///////////////////////////////////////////////////////////////////////////////
        # Shared memory layout: dQ
        # ///////////////////////////////////////////////////////////////////////////////
        # We can't just use kHeadDim here. E.g. if MMA shape is 64 x 96 but split across 2 WGs,
        # then setting kBlockKSmem to 32 will cause "Static shape_div failure".
        # We want to treat it as 64 x 48, so kBlockKSmem should be 16.
        # TODO: this is hard-coded for hdim 128
        self.sdQ_layout = sm100_utils_basic.make_smem_layout_epi(self.dtype, LayoutEnum.ROW_MAJOR, (self.tile_m, self.tile_hdim), 1)

    @cute.jit
    def __call__(
        self,
        mdQaccum: cute.Tensor,
        mdQ: cute.Tensor,
        scale: cutlass.Float32,
        mCuSeqlensQ: Optional[cute.Tensor],
        mSeqUsedQ: Optional[cute.Tensor],
        # Always keep stream as the last parameter (EnvStream: obtained implicitly via TVM FFI).
        stream: cuda.CUstream = None,
    ):
        # Get the data type and check if it is fp16 or bf16
        if const_expr(mdQ.element_type not in [cutlass.Float16, cutlass.BFloat16]):
            raise TypeError("Only Float16 or BFloat16 is supported")
        if const_expr(mdQaccum is not None):
            if const_expr(mdQaccum.element_type not in [cutlass.Float32]):
                raise TypeError("dQaccum tensor must be Float32")

        mdQaccum, mdQ = [assume_tensor_aligned(t) for t in (mdQaccum, mdQ)]

        self.tiled_mma = self._get_tiled_mma()
        self._setup_attributes()

        smem_size = max(
            cute.size_in_bytes(cutlass.Float32, self.sdQaccum_layout),
            cute.size_in_bytes(self.dtype, self.sdQ_layout),
        )

        if const_expr(mCuSeqlensQ is not None):
            TileScheduler = SingleTileVarlenScheduler
            num_head = mdQ.shape[1]
            num_batch = mCuSeqlensQ.shape[0] - 1
            num_block = cute.ceil_div(mdQ.shape[0], self.tile_m)
        else:
            TileScheduler = SingleTileScheduler
            num_head = mdQ.shape[2]
            num_batch = mdQ.shape[0]
            num_block = cute.ceil_div(mdQ.shape[1], self.tile_m)

        tile_sched_args = TileSchedulerArguments(
            num_block=num_block,
            num_head=num_head,
            num_batch=num_batch,
            num_splits=1,
            seqlen_k=0,
            headdim=mdQ.shape[2],
            headdim_v=0,
            total_q=mdQ.shape[0],
            tile_shape_mn=(self.tile_m, 1),
            mCuSeqlensQ=mCuSeqlensQ,
            mSeqUsedQ=mSeqUsedQ,
        )

        tile_sched_params = TileScheduler.to_underlying_arguments(tile_sched_args)
        grid_dim = TileScheduler.get_grid_shape(tile_sched_params)

        # grid_dim: (m_block, num_head, batch_size)
        self.kernel(
            mdQaccum,
            mdQ,
            mCuSeqlensQ,
            mSeqUsedQ,
            scale,
            self.tiled_mma,
            self.sdQaccum_layout,
            self.sdQ_layout,
            self.g2s_tiled_copy_dQaccum,
            self.s2r_tiled_copy_dQaccum,
            self.gmem_tiled_copy_dQ,
            tile_sched_params,
            TileScheduler,
        ).launch(
            grid=grid_dim,
            block=[self.num_threads, 1, 1],
            smem=smem_size,
            stream=stream,
        )

    @cute.kernel
    def kernel(
        self,
        mdQaccum: cute.Tensor,
        mdQ: cute.Tensor,
        mCuSeqlensQ: Optional[cute.Tensor],
        mSeqUsedQ: Optional[cute.Tensor],
        scale: cutlass.Float32,
        tiled_mma: cute.TiledMma,
        sdQaccum_layout: cute.Layout,
        sdQ_layout: cute.ComposedLayout,
        g2s_tiled_copy_dQaccum: cute.TiledCopy,
        s2r_tiled_copy_dQaccum: cute.TiledCopy,
        gmem_tiled_copy_dQ: cute.TiledCopy,
        tile_sched_params: ParamsBase,
        TileScheduler: cutlass.Constexpr[Callable],
    ):
        # ///////////////////////////////////////////////////////////////////////////////
        # Get shared memory buffer
        # ///////////////////////////////////////////////////////////////////////////////
        smem = cutlass.utils.SmemAllocator()
        sdQaccum = smem.allocate_tensor(cutlass.Float32, sdQaccum_layout, byte_alignment=1024)
        sdQaccum_flat = cute.make_tensor(sdQaccum.iterator, cute.make_layout(cute.size(sdQaccum)))
        # extra stage dimension
        sdQ = cute.make_tensor(
            cute.recast_ptr(sdQaccum.iterator, sdQ_layout.inner, dtype=self.dtype),
            sdQ_layout.outer,
        )[None, None, 0]
        sdQt = layout_utils.transpose_view(sdQ)

        # Thread index, block index
        tidx, _, _ = cute.arch.thread_idx()

        tile_scheduler = TileScheduler.create(tile_sched_params)
        work_tile = tile_scheduler.initial_work_tile_info()

        m_block, head_idx, batch_idx, _ = work_tile.tile_idx

        if work_tile.is_valid_tile:
            # ///////////////////////////////////////////////////////////////////////////////
            # Get the appropriate tiles for this thread block.
            # ///////////////////////////////////////////////////////////////////////////////

            seqlen = SeqlenInfoQK.create(
                batch_idx,
                mdQ.shape[1],
                0,
                mCuSeqlensQ=mCuSeqlensQ,
                mCuSeqlensK=None,
                mSeqUsedQ=mSeqUsedQ,
                mSeqUsedK=None,
                tile_m=self.tile_m,
            )
            if const_expr(not seqlen.has_cu_seqlens_q):
                mdQ_cur = mdQ[batch_idx, None, head_idx, None]
                mdQaccum_cur = mdQaccum[batch_idx, head_idx, None]
                head_dim = mdQ.shape[3]
            else:
                padded_offset_q = seqlen.padded_offset_q
                mdQ_cur = cute.domain_offset((seqlen.offset_q, 0), mdQ[None, head_idx, None])
                mdQaccum_cur = cute.domain_offset((padded_offset_q * self.tile_hdim,), mdQaccum[head_idx, None])
                head_dim = mdQ.shape[2]

                # NOTE: Compiler doesn't seem to recognize that padding
                # by padded_offset_q * self.tile_hdim keeps alignment
                # since statically divisible by 4

                mdQaccum_cur_ptr = cute.make_ptr(
                    dtype=mdQaccum_cur.element_type,
                    value=mdQaccum_cur.iterator.toint(),
                    mem_space=mdQaccum_cur.iterator.memspace,
                    assumed_align=mdQaccum.iterator.alignment,
                )
                mdQaccum_cur = cute.make_tensor(mdQaccum_cur_ptr, mdQaccum_cur.layout)

            gdQaccum = cute.local_tile(mdQaccum_cur, (self.tile_m * self.tile_hdim,), (m_block,))
            gdQ = cute.local_tile(mdQ_cur, (self.tile_m, self.tile_hdim), (m_block, 0))

            seqlen_q = seqlen.seqlen_q
            seqlen_q_rounded = cute.round_up(seqlen_q, self.tile_m)

            # Step 1: load dQaccum from gmem to smem
            g2s_thr_copy_dQaccum = g2s_tiled_copy_dQaccum.get_slice(tidx)
            tdQgdQaccum = g2s_thr_copy_dQaccum.partition_S(gdQaccum)
            tdQsdQaccumg2s = g2s_thr_copy_dQaccum.partition_D(sdQaccum_flat)
            cute.copy(g2s_tiled_copy_dQaccum, tdQgdQaccum, tdQsdQaccumg2s)
            cute.arch.cp_async_commit_group()
            cute.arch.cp_async_wait_group(0)
            cute.arch.barrier()

            # Step 2: load dQ from smem to rmem
            s2r_thr_copy_dQaccum = s2r_tiled_copy_dQaccum.get_slice(tidx)
            tdQsdQaccum = s2r_thr_copy_dQaccum.partition_S(sdQaccum)
            thr_mma = tiled_mma.get_slice(0)
            dQacc_shape = tiled_mma.partition_shape_C((self.tile_m, self.tile_hdim))
            tdQtdQ = tiled_mma.make_fragment_C(dQacc_shape)
            tdQcdQ = thr_mma.partition_C(cute.make_identity_tensor((self.tile_m, self.tile_hdim)))
            tmem_load_atom = cute.make_copy_atom(
                tcgen05.copy.Ld32x32bOp(tcgen05.copy.Repetition(self.dQ_reduce_ncol)),
                Float32,
            )
            tiled_copy_t2r = tcgen05.make_tmem_copy(tmem_load_atom, tdQtdQ)
            thr_copy_t2r = tiled_copy_t2r.get_slice(tidx)
            tdQrdQ_t2r_shape = thr_copy_t2r.partition_D(tdQcdQ).shape
            acc = cute.make_rmem_tensor(tdQrdQ_t2r_shape, Float32)
            tdQrdQaccum = cute.make_tensor(acc.iterator, cute.make_layout(tdQsdQaccum.shape))
            cute.autovec_copy(tdQsdQaccum, tdQrdQaccum)
            # Convert tdQrdQaccum from fp32 to fp16/bf16
            rdQ = cute.make_rmem_tensor_like(acc, self.dtype)
            rdQ.store((acc.load() * scale).to(self.dtype))

            # Step 3: Copy dQ from register to smem
            cute.arch.barrier()  # make sure all threads have finished loading dQaccum
            thr_layout_r2s_dQ = cute.make_layout((self.num_threads, 1))  # 128 threads
            val_layout_r2s_dQ = cute.make_layout((1, 128 // self.dtype.width))
            copy_atom_r2s_dQ = cute.make_copy_atom(
                cute.nvgpu.CopyUniversalOp(),
                self.dtype,
                num_bits_per_copy=128,
            )
            tiled_copy_r2s_dQ = cute.make_tiled_copy_tv(copy_atom_r2s_dQ, thr_layout_r2s_dQ, val_layout_r2s_dQ)
            thr_copy_r2s_dQ = tiled_copy_r2s_dQ.get_slice(tidx)
            cdQ = cute.make_identity_tensor((self.tile_m, self.tile_hdim))
            taccdQcdQ_shape = thr_copy_r2s_dQ.partition_S(cdQ).shape
            taccdQrdQ = cute.make_tensor(rdQ.iterator, taccdQcdQ_shape)
            taccdQsdQ = thr_copy_r2s_dQ.partition_D(sdQ if const_expr(not self.dQ_swapAB) else sdQt)
            cute.copy(thr_copy_r2s_dQ, taccdQrdQ, taccdQsdQ)

            # Step 4: Copy dQ from smem to register to prepare for coalesced write to gmem
            cute.arch.barrier()  # make sure all smem stores are done
            gmem_thr_copy_dQ = gmem_tiled_copy_dQ.get_slice(tidx)
            tdQgdQ = gmem_thr_copy_dQ.partition_S(gdQ)
            tdQsdQ = gmem_thr_copy_dQ.partition_D(sdQ)
            tdQrdQ = cute.make_rmem_tensor_like(tdQsdQ, self.dtype)
            # TODO: check OOB when reading from smem if kBlockM isn't evenly tiled
            cute.autovec_copy(tdQsdQ, tdQrdQ)

            # Step 5: Copy dQ from register to gmem
            tdQcdQ = gmem_thr_copy_dQ.partition_S(cdQ)
            tdQpdQ = utils.predicate_k(tdQcdQ, limit=head_dim)
            for rest_m in cutlass.range(cute.size(tdQrdQ.shape[1]), unroll_full=True):
                if tdQcdQ[0, rest_m, 0][0] < seqlen_q - m_block * self.tile_m:
                    cute.copy(
                        gmem_tiled_copy_dQ,
                        tdQrdQ[None, rest_m, None],
                        tdQgdQ[None, rest_m, None],
                        pred=tdQpdQ[None, rest_m, None],
                    )



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Optional
from dataclasses import dataclass

import cutlass
import cutlass.cute as cute
from cutlass import Int32, const_expr

from cudnn.block_sparse_attention.csrc.utils import copy_utils

"""
This consolidates all the info related to sequence length. This is so that we can do all
the gmem reads once at the beginning of each tile, rather than having to repeat these reads
to compute various things like n_block_min, n_block_max, etc.
"""


@dataclass(frozen=True)
class SeqlenInfo:
    offset: Int32
    offset_padded: Int32
    seqlen: Int32
    has_cu_seqlens: cutlass.Constexpr[bool] = False

    @staticmethod
    def create(
        batch_idx: Int32,
        seqlen_static: Int32,
        cu_seqlens: Optional[cute.Tensor] = None,
        seqused: Optional[cute.Tensor] = None,
        tile: cutlass.Constexpr[int] = 128,
    ):
        offset = 0 if const_expr(cu_seqlens is None) else cu_seqlens[batch_idx]
        offset_padded = (
            0
            if const_expr(cu_seqlens is None)
            # Add divby so that the compiler knows the alignment when moving by offset_padded
            else cute.assume((offset + batch_idx * tile) // tile * tile, divby=tile)
        )
        if const_expr(seqused is not None):
            seqlen = seqused[batch_idx]
        elif const_expr(cu_seqlens is not None):
            seqlen = cu_seqlens[batch_idx + 1] - cu_seqlens[batch_idx]
        else:
            seqlen = seqlen_static
        return SeqlenInfo(offset, offset_padded, seqlen, has_cu_seqlens=cu_seqlens is not None)

    def offset_batch(
        self,
        mT: cute.Tensor,
        batch_idx: Int32,
        dim: int,
        padded: cutlass.Constexpr[bool] = False,
        multiple: int = 1,
    ) -> cute.Tensor:
        """Offset a tensor by batch index. batch dim is at position `dim`, seqlen is at dim=0."""
        if const_expr(not self.has_cu_seqlens):
            idx = (None,) * dim + (batch_idx,) + (None,) * (cute.rank(mT) - 1 - dim)
            return mT[idx]
        else:
            off = multiple * (self.offset if const_expr(not padded) else self.offset_padded)
            offset = off if const_expr(cute.rank(mT.shape[0]) == 1) else (0, off)
            idx = (offset,) + (None,) * (cute.rank(mT) - 1)
            return cute.domain_offset(idx, mT)


@dataclass(frozen=True)
class SeqlenInfoQK:
    offset_q: Int32
    offset_k: Int32
    padded_offset_q: Int32
    padded_offset_k: Int32
    seqlen_q: Int32
    seqlen_k: Int32
    m_block_offset: Int32
    block_idx_offset: Int32
    num_n_blocks: Int32
    has_cu_seqlens_q: cutlass.Constexpr[bool]
    has_cu_seqlens_k: cutlass.Constexpr[bool]
    has_seqused_q: cutlass.Constexpr[bool]
    has_seqused_k: cutlass.Constexpr[bool]

    @staticmethod
    def create(
        batch_idx: Int32,
        seqlen_q_static: Int32,
        seqlen_k_static: Int32,
        mCuSeqlensQ: Optional[cute.Tensor] = None,
        mCuSeqlensK: Optional[cute.Tensor] = None,
        mSeqUsedQ: Optional[cute.Tensor] = None,
        mSeqUsedK: Optional[cute.Tensor] = None,
        mCuTotalMBlocks: Optional[cute.Tensor] = None,
        mCuBlockIdxOffsets: Optional[cute.Tensor] = None,
        tile_m: cutlass.Constexpr[Int32] = 128,
        tile_n: cutlass.Constexpr[Int32] = 128,
    ):
        offset_q = 0 if const_expr(mCuSeqlensQ is None) else mCuSeqlensQ[batch_idx]
        offset_k = 0 if const_expr(mCuSeqlensK is None) else mCuSeqlensK[batch_idx]
        padded_offset_q = 0 if const_expr(mCuSeqlensQ is None) else cute.assume((offset_q + batch_idx * tile_m) // tile_m * tile_m, divby=tile_m)
        padded_offset_k = 0 if const_expr(mCuSeqlensK is None) else cute.assume((offset_k + batch_idx * tile_n) // tile_n * tile_n, divby=tile_n)
        if const_expr(mSeqUsedQ is not None):
            seqlen_q = mSeqUsedQ[batch_idx]
        else:
            seqlen_q = seqlen_q_static if const_expr(mCuSeqlensQ is None) else mCuSeqlensQ[batch_idx + 1] - offset_q
        if const_expr(mSeqUsedK is not None):
            seqlen_k = mSeqUsedK[batch_idx]
        else:
            seqlen_k = seqlen_k_static if const_expr(mCuSeqlensK is None) else mCuSeqlensK[batch_idx + 1] - offset_k
        m_block_offset = 0 if const_expr(mCuTotalMBlocks is None) else mCuTotalMBlocks[batch_idx]
        num_n_blocks = (seqlen_k + tile_n - 1) // tile_n
        block_idx_offset = mCuBlockIdxOffsets[batch_idx] if const_expr(mCuBlockIdxOffsets is not None) else m_block_offset * num_n_blocks
        return SeqlenInfoQK(
            offset_q,
            offset_k,
            padded_offset_q,
            padded_offset_k,
            seqlen_q,
            seqlen_k,
            m_block_offset,
            block_idx_offset,
            num_n_blocks,
            has_cu_seqlens_q=mCuSeqlensQ is not None,
            has_cu_seqlens_k=mCuSeqlensK is not None,
            has_seqused_q=mSeqUsedQ is not None,
            has_seqused_k=mSeqUsedK is not None,
        )

    def offset_batch_Q(
        self,
        mQ: cute.Tensor,
        batch_idx: Int32,
        dim: int,
        padded: cutlass.Constexpr[bool] = False,
        ragged: cutlass.Constexpr[bool] = False,
    ) -> cute.Tensor:
        """Seqlen must be the first dimension of mQ"""
        if const_expr(not ragged):
            if const_expr(not self.has_cu_seqlens_q):
                idx = (None,) * dim + (batch_idx,) + (None,) * (cute.rank(mQ) - 1 - dim)
                return mQ[idx]
            else:
                offset_q = self.offset_q if const_expr(not padded) else self.padded_offset_q
                offset_q = offset_q if const_expr(cute.rank(mQ.shape[0]) == 1) else (None, offset_q)
                idx = (offset_q,) + (None,) * (cute.rank(mQ) - 1)
                return cute.domain_offset(idx, mQ)
        else:
            if const_expr(not self.has_cu_seqlens_q):
                offset_q = 0
                idx = (None,) * dim + (batch_idx,) + (None,) * (cute.rank(mQ) - 1 - dim)
                mQ = mQ[idx]
            else:
                offset_q = self.offset_q if const_expr(not padded) else self.padded_offset_q
            if const_expr(cute.rank(mQ.shape[0]) == 1):
                return copy_utils.offset_ragged_tensor(mQ, offset_q, self.seqlen_q, ragged_dim=0, ptr_shift=True)
            else:  # PackGQA
                assert cute.rank(mQ.shape[0]) == 2
                # Unpack before calling offset_ragged_tensor, then pack
                idx = ((None, None),) + (None,) * (cute.rank(mQ) - 1)
                mQ = mQ[idx]
                mQ = copy_utils.offset_ragged_tensor(mQ, offset_q, self.seqlen_q, ragged_dim=1, ptr_shift=True)
                return cute.group_modes(mQ, 0, 2)

    def offset_batch_K(
        self,
        mK: cute.Tensor,
        batch_idx: Int32,
        dim: int,
        padded: cutlass.Constexpr[bool] = False,
        ragged: cutlass.Constexpr[bool] = False,
        multiple: int = 1,
    ) -> cute.Tensor:
        """Seqlen must be the first dimension of mK"""
        if const_expr(not ragged):
            if const_expr(not self.has_cu_seqlens_k):
                idx = (None,) * dim + (batch_idx,) + (None,) * (cute.rank(mK) - 1 - dim)
                return mK[idx]
            else:
                offset_k = self.offset_k if const_expr(not padded) else self.padded_offset_k
                offset_k *= multiple
                idx = (offset_k,) + (None,) * (cute.rank(mK) - 1)
                return cute.domain_offset(idx, mK)
        else:
            if const_expr(not self.has_cu_seqlens_k):
                offset_k = 0
                idx = (None,) * dim + (batch_idx,) + (None,) * (cute.rank(mK) - 1 - dim)
                mK = mK[idx]
            else:
                offset_k = self.offset_k if const_expr(not padded) else self.padded_offset_k
                offset_k *= multiple
            return copy_utils.offset_ragged_tensor(mK, offset_k, self.seqlen_k, ragged_dim=0, ptr_shift=True)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""Host-side MXFP8 scale packing helpers for the SM100 Indexer DSL path."""

from __future__ import annotations

import math

import torch

from cudnn.deepseek_sparse_attention.utils.runtime import ceil_div as _ceil_div

_ATOM_MN = 128
_ATOM_K = 4
_DATA_PATHS = 32
_DATA_PATH_STRIDE = 16
_ATOM_ELEMS = 512


def _cu_seqlens_to_list(
    cu_seqlens: torch.Tensor,
    *,
    total: int,
    name: str,
) -> list[int]:
    if cu_seqlens.ndim != 1:
        raise ValueError(f"{name} must be 1D")
    values = [int(v) for v in cu_seqlens.detach().cpu().tolist()]
    if not values:
        raise ValueError(f"{name} must have at least one element")
    if values[0] != 0:
        raise ValueError(f"{name}[0] must be 0")
    if values[-1] != total:
        raise ValueError(f"{name}[-1] must equal the scale tensor sequence length")
    if any(cur < prev for prev, cur in zip(values, values[1:])):
        raise ValueError(f"{name} must be nondecreasing")
    return values


def make_scale_cu_seqlens_padded(
    cu_seqlens: torch.Tensor,
    token_alignment: int,
) -> torch.Tensor:
    """Return per-sequence padded token prefixes for a compact THD tensor.

    ``cu_seqlens`` remains the prefix for compact Q/K data.  The returned
    ``(B+1,)`` int32 tensor is the prefix for the independently padded MXFP8
    scale storage, where each sequence length is minimally rounded up to
    ``token_alignment``. Callers may instead pass the THD scale packers any span
    that covers the logical sequence and is a multiple of ``token_alignment``.
    The operation stays on the input device.
    """
    if cu_seqlens.ndim != 1 or cu_seqlens.numel() < 2:
        raise ValueError("cu_seqlens must be 1D with at least two elements")
    if cu_seqlens.dtype != torch.int32:
        raise ValueError("cu_seqlens must be int32")
    if token_alignment <= 0:
        raise ValueError("token_alignment must be positive")
    lengths = cu_seqlens[1:] - cu_seqlens[:-1]
    padded_lengths = ((lengths + int(token_alignment) - 1) // int(token_alignment)) * int(token_alignment)
    out = torch.zeros_like(cu_seqlens)
    torch.cumsum(padded_lengths, dim=0, out=out[1:])
    return out


def _validate_scale_cu_seqlens_padded(
    cu_seqlens: list[int],
    cu_seqlens_scale_padded: torch.Tensor,
    *,
    token_alignment: int,
    name: str,
) -> list[int]:
    if cu_seqlens_scale_padded.ndim != 1:
        raise ValueError(f"{name} must be 1D")
    if cu_seqlens_scale_padded.dtype != torch.int32:
        raise ValueError(f"{name} must be int32")
    values = [int(v) for v in cu_seqlens_scale_padded.detach().cpu().tolist()]
    if len(values) != len(cu_seqlens):
        raise ValueError(f"{name} must have the same length as cu_seqlens")
    if not values or values[0] != 0:
        raise ValueError(f"{name}[0] must be 0")
    if any(cur < prev for prev, cur in zip(values, values[1:])):
        raise ValueError(f"{name} must be nondecreasing")
    for b, (logical0, logical1, padded0, padded1) in enumerate(zip(cu_seqlens, cu_seqlens[1:], values, values[1:])):
        logical_len = logical1 - logical0
        padded_len = padded1 - padded0
        if padded_len < logical_len or padded_len % token_alignment != 0:
            raise ValueError(f"{name} batch {b} span must be at least {logical_len} and a " f"multiple of {token_alignment}, got {padded_len}")
    return values


def _blockscaled_indices(
    mn: int,
    sf_groups: int,
    l_size: int,
    *,
    device: torch.device,
) -> torch.Tensor:
    """Return flat indices for NVIDIA's packed blockscaled SF layout."""
    mn_tiles = _ceil_div(mn, _ATOM_MN)
    sf_tiles = _ceil_div(sf_groups, _ATOM_K)
    sf_per_l = _ATOM_ELEMS * mn_tiles * sf_tiles

    m, sf = torch.meshgrid(
        torch.arange(mn, device=device, dtype=torch.int64),
        torch.arange(sf_groups, device=device, dtype=torch.int64),
        indexing="ij",
    )
    base = (
        (m // _ATOM_MN) * (_ATOM_ELEMS * sf_tiles)
        + (sf // _ATOM_K) * _ATOM_ELEMS
        + _DATA_PATH_STRIDE * (m % _DATA_PATHS)
        + _ATOM_K * ((m % _ATOM_MN) // _DATA_PATHS)
        + (sf % _ATOM_K)
    )
    l_offsets = torch.arange(l_size, device=device, dtype=torch.int64)[:, None, None]
    return base.unsqueeze(0) + l_offsets * sf_per_l


def pack_blockscaled_scale_mkl(scale_mkl: torch.Tensor) -> torch.Tensor:
    """Pack logical ``(MN, sf_groups, L)`` scales into Blackwell SF layout.

    The returned tensor has physical shape ``(L, padded_MN, padded_sf_groups)``.
    ``sf_groups`` is ``ceil_div(head_dim, 32)`` for MXFP8.
    """
    if scale_mkl.ndim != 3:
        raise ValueError(f"scale_mkl must have shape (MN, sf_groups, L), got {tuple(scale_mkl.shape)}")
    mn, sf_groups, l_size = scale_mkl.shape
    m_padded = _ceil_div(mn, _ATOM_MN) * _ATOM_MN
    sf_padded = _ceil_div(sf_groups, _ATOM_K) * _ATOM_K
    packed = torch.zeros(
        (l_size, m_padded, sf_padded),
        device=scale_mkl.device,
        dtype=scale_mkl.dtype,
    )
    idx = _blockscaled_indices(mn, sf_groups, l_size, device=scale_mkl.device)
    packed.reshape(-1)[idx.reshape(-1)] = scale_mkl.permute(2, 0, 1).reshape(-1)
    return packed.contiguous()


def unpack_blockscaled_scale_mkl(
    scale_packed: torch.Tensor,
    mn: int,
    sf_groups: int,
) -> torch.Tensor:
    """Unpack a packed scale tensor back to logical ``(MN, sf_groups, L)``."""
    if scale_packed.ndim != 3:
        raise ValueError("scale_packed must have shape (L, padded_MN, padded_sf_groups), " f"got {tuple(scale_packed.shape)}")
    l_size = scale_packed.shape[0]
    idx = _blockscaled_indices(mn, sf_groups, l_size, device=scale_packed.device)
    return scale_packed.reshape(-1)[idx].permute(1, 2, 0).contiguous()


def logical_q_scale_to_mkl(
    q_scale: torch.Tensor,
    qhead_per_kv_head: int,
) -> torch.Tensor:
    """Convert BSHG Q scales to packed-Q logical ``(MN, sf_groups, L)`` order.

    Row order matches the kernel PackGQA order:
    ``packed_m = q_token * qhead_per_kv_head + h_local``.
    The L dimension is flattened as ``batch * n_heads_kv + kv_head``.
    """
    if q_scale.ndim != 4:
        raise ValueError(f"q_scale must have shape (B, S_q, H_q, sf_groups), got {tuple(q_scale.shape)}")
    if qhead_per_kv_head <= 0:
        raise ValueError("qhead_per_kv_head must be positive")
    bs, seqlen_q, n_heads_q, sf_groups = q_scale.shape
    if n_heads_q % qhead_per_kv_head != 0:
        raise ValueError(f"n_heads_q ({n_heads_q}) must be divisible by qhead_per_kv_head " f"({qhead_per_kv_head})")
    n_heads_kv = n_heads_q // qhead_per_kv_head
    q_view = q_scale.view(bs, seqlen_q, n_heads_kv, qhead_per_kv_head, sf_groups)
    return q_view.permute(1, 3, 4, 0, 2).reshape(seqlen_q * qhead_per_kv_head, sf_groups, bs * n_heads_kv).contiguous()


def logical_k_scale_to_mkl(k_scale: torch.Tensor) -> torch.Tensor:
    """Convert BSHG K scales to logical ``(MN, sf_groups, L)`` order."""
    if k_scale.ndim != 4:
        raise ValueError(f"k_scale must have shape (B, S_k, H_kv, sf_groups), got {tuple(k_scale.shape)}")
    bs, seqlen_k, n_heads_kv, sf_groups = k_scale.shape
    return k_scale.permute(1, 3, 0, 2).reshape(seqlen_k, sf_groups, bs * n_heads_kv).contiguous()


def logical_q_scale_to_mkl_thd(
    q_scale: torch.Tensor,
    cu_seqlens_q: torch.Tensor,
    cu_seqlens_q_scale_padded: torch.Tensor,
    qhead_per_kv_head: int,
) -> torch.Tensor:
    """Convert compact THG Q scales to padded packed-Q ``(MN, groups, L)``.

    Sequences are concatenated along ``MN`` using
    ``cu_seqlens_q_scale_padded``. Each sequence's token span must cover the
    logical sequence and, after multiplying by ``qhead_per_kv_head``, produce
    a multiple of 128 packed-MN rows. The caller-provided span is preserved
    exactly. ``L`` contains only KV heads, not the batch dimension.
    """
    if q_scale.ndim != 3:
        raise ValueError(f"q_scale must have shape (total_q, H_q, sf_groups), got {tuple(q_scale.shape)}")
    if qhead_per_kv_head <= 0:
        raise ValueError("qhead_per_kv_head must be positive")
    total_q, n_heads_q, sf_groups = q_scale.shape
    if n_heads_q % qhead_per_kv_head != 0:
        raise ValueError(f"n_heads_q ({n_heads_q}) must be divisible by qhead_per_kv_head " f"({qhead_per_kv_head})")
    cu_q = _cu_seqlens_to_list(cu_seqlens_q, total=total_q, name="cu_seqlens_q")
    n_heads_kv = n_heads_q // qhead_per_kv_head
    token_alignment = 128 // math.gcd(128, qhead_per_kv_head)
    cu_q_scale = _validate_scale_cu_seqlens_padded(
        cu_q,
        cu_seqlens_q_scale_padded,
        token_alignment=token_alignment,
        name="cu_seqlens_q_scale_padded",
    )
    bs = len(cu_q) - 1
    out = torch.zeros(
        cu_q_scale[-1] * qhead_per_kv_head,
        sf_groups,
        n_heads_kv,
        device=q_scale.device,
        dtype=q_scale.dtype,
    )
    for b in range(bs):
        q0 = cu_q[b]
        q1 = cu_q[b + 1]
        q_len = q1 - q0
        if q_len == 0:
            continue
        q_view = q_scale[q0:q1].view(q_len, n_heads_kv, qhead_per_kv_head, sf_groups)
        q_scale0 = cu_q_scale[b] * qhead_per_kv_head
        out[q_scale0 : q_scale0 + q_len * qhead_per_kv_head] = q_view.permute(0, 2, 3, 1).reshape(q_len * qhead_per_kv_head, sf_groups, n_heads_kv)
    return out.contiguous()


def logical_k_scale_to_mkl_thd(
    k_scale: torch.Tensor,
    cu_seqlens_k: torch.Tensor,
    cu_seqlens_k_scale_padded: torch.Tensor,
) -> torch.Tensor:
    """Convert compact THG K scales to padded logical ``(MN, groups, L)``."""
    if k_scale.ndim != 3:
        raise ValueError(f"k_scale must have shape (total_k, H_kv, sf_groups), got {tuple(k_scale.shape)}")
    total_k, n_heads_kv, sf_groups = k_scale.shape
    cu_k = _cu_seqlens_to_list(cu_seqlens_k, total=total_k, name="cu_seqlens_k")
    cu_k_scale = _validate_scale_cu_seqlens_padded(
        cu_k,
        cu_seqlens_k_scale_padded,
        token_alignment=128,
        name="cu_seqlens_k_scale_padded",
    )
    bs = len(cu_k) - 1
    out = torch.zeros(
        cu_k_scale[-1],
        sf_groups,
        n_heads_kv,
        device=k_scale.device,
        dtype=k_scale.dtype,
    )
    for b in range(bs):
        k0 = cu_k[b]
        k1 = cu_k[b + 1]
        k_len = k1 - k0
        if k_len == 0:
            continue
        k_scale0 = cu_k_scale[b]
        out[k_scale0 : k_scale0 + k_len] = k_scale[k0:k1].permute(0, 2, 1)
    return out.contiguous()


def pack_q_scale_bshd(q_scale: torch.Tensor, qhead_per_kv_head: int = 64) -> torch.Tensor:
    """Pack logical BSHD Q scales for the Indexer MXFP8 kernel."""
    return pack_blockscaled_scale_mkl(logical_q_scale_to_mkl(q_scale, qhead_per_kv_head))


def pack_k_scale_bshd(k_scale: torch.Tensor) -> torch.Tensor:
    """Pack logical BSHD K scales for the Indexer MXFP8 kernel."""
    return pack_blockscaled_scale_mkl(logical_k_scale_to_mkl(k_scale))


def pack_q_scale_thd(
    q_scale: torch.Tensor,
    cu_seqlens_q: torch.Tensor,
    cu_seqlens_q_scale_padded: torch.Tensor,
    qhead_per_kv_head: int = 64,
) -> torch.Tensor:
    """Pack logical THD Q scales for the Indexer MXFP8 kernel."""
    return pack_blockscaled_scale_mkl(
        logical_q_scale_to_mkl_thd(
            q_scale,
            cu_seqlens_q,
            cu_seqlens_q_scale_padded,
            qhead_per_kv_head,
        )
    )


def pack_k_scale_thd(
    k_scale: torch.Tensor,
    cu_seqlens_k: torch.Tensor,
    cu_seqlens_k_scale_padded: torch.Tensor,
) -> torch.Tensor:
    """Pack logical THD K scales for the Indexer MXFP8 kernel."""
    return pack_blockscaled_scale_mkl(
        logical_k_scale_to_mkl_thd(
            k_scale,
            cu_seqlens_k,
            cu_seqlens_k_scale_padded,
        )
    )



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from .api import (
    IndexerForward,
    indexer_forward_wrapper,
    indexer_forward_top_k_wrapper,
)
from ._compressed_top_k_sm100 import (
    compress_topk_cand_buffer_size,
    compress_topk_cand_buffer_size_thd,
)

__all__ = [
    "IndexerForward",
    "indexer_forward_wrapper",
    "indexer_forward_top_k_wrapper",
    "compress_topk_cand_buffer_size",
    "compress_topk_cand_buffer_size_thd",
]



# Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT
"""SMEM FWHT device functions for the fused RHT output (orthonormal Sylvester H16).

Split into LOAD helpers (sD -> registers) and rmem COMPUTE functions so the kernel
can signal "sD consumed" (lockstep-barrier arrive) right after the loads — the
FWHT/quant compute and all stores then overlap the ACT warps' next subtile."""

import cutlass
import cutlass.cute as cute

from .moe_kernel_helpers import fmax, fmin
from .quant_utils import (
    _nvfp4_quant_row,
    get_dtype_rcp_limits,
)

HADAMARD_SIZE = 16


@cute.jit
def load_colwise_pairs_bf16(sD, d_buffer, tidx):
    """16 vectorized 4-byte pair loads of the thread's TWO adjacent feature columns
    of one 16-token block, interleaved token-major: rmem[2*i + c] =
    sD(token_base + i, feature + c)."""
    token_block = tidx // HADAMARD_SIZE
    feat_pair = tidx % HADAMARD_SIZE
    token_base = token_block * HADAMARD_SIZE

    rmem_bf16 = cute.make_rmem_tensor((2 * HADAMARD_SIZE,), cutlass.BFloat16)
    rmem_pairs = cute.zipped_divide(rmem_bf16, (2,))
    sD_pairs = cute.zipped_divide(cute.slice_(sD, (None, None, d_buffer)), (1, 2))
    for i in cutlass.range_constexpr(HADAMARD_SIZE):
        cute.autovec_copy(
            cute.slice_(sD_pairs, ((None, None), (token_base + i, feat_pair))),
            cute.slice_(rmem_pairs, ((None,), i)),
        )
    return rmem_bf16


@cute.jit
def _colwise_fwht_inplace(tCompute):
    """In-place natural-order 16-point FWHT over both interleaved token-major
    columns (tCompute[2*i + c] = column c, token i)."""
    for c in cutlass.range_constexpr(2):
        for base in cutlass.range_constexpr(0, HADAMARD_SIZE, 2):
            x = tCompute[2 * base + c]
            y = tCompute[2 * (base + 1) + c]
            tCompute[2 * base + c] = x + y
            tCompute[2 * (base + 1) + c] = x - y
        for base in cutlass.range_constexpr(0, HADAMARD_SIZE, 4):
            for i in cutlass.range_constexpr(2):
                x = tCompute[2 * (base + i) + c]
                y = tCompute[2 * (base + i + 2) + c]
                tCompute[2 * (base + i) + c] = x + y
                tCompute[2 * (base + i + 2) + c] = x - y
        for base in cutlass.range_constexpr(0, HADAMARD_SIZE, 8):
            for i in cutlass.range_constexpr(4):
                x = tCompute[2 * (base + i) + c]
                y = tCompute[2 * (base + i + 4) + c]
                tCompute[2 * (base + i) + c] = x + y
                tCompute[2 * (base + i + 4) + c] = x - y
        for i in cutlass.range_constexpr(8):
            x = tCompute[2 * i + c]
            y = tCompute[2 * (i + 8) + c]
            tCompute[2 * i + c] = x + y
            tCompute[2 * (i + 8) + c] = x - y


@cute.jit
def hadamard_rmem_colwise_fwht(rmem_bf16, d_buffer, tidx, sRht):
    """Per-feature bf16 FWHT over 16-token blocks from the load_colwise_pairs_bf16
    registers, stored to sRht at the SAME (token, feature) coords the input was
    read from as one 4-byte pair store per token row.
    (The colwise QUANT path lives in hadamard_rmem_colwise_fwht_quant.)

    In-place natural-order FWHT => output i is the transformed value for token
    token_base+i, matching torch_rht's block layout."""
    token_block = tidx // HADAMARD_SIZE
    feat_pair = tidx % HADAMARD_SIZE
    token_base = token_block * HADAMARD_SIZE

    n_vals = 2 * HADAMARD_SIZE
    tCompute = cute.make_rmem_tensor((n_vals,), cutlass.Float32)
    for i in cutlass.range_constexpr(n_vals):
        tCompute[i] = rmem_bf16[i].to(cutlass.Float32)

    _colwise_fwht_inplace(tCompute)

    # Orthonormal scale, rounded through bf16 (== the bf16 output values).
    for i in cutlass.range_constexpr(n_vals):
        tCompute[i] = (tCompute[i] * cutlass.Float32(0.25)).to(cutlass.BFloat16).to(cutlass.Float32)

    rmem_st = cute.make_rmem_tensor((2,), cutlass.BFloat16)
    sRht_pairs = cute.zipped_divide(cute.slice_(sRht, (None, None, d_buffer)), (1, 2))
    for i in cutlass.range_constexpr(HADAMARD_SIZE):
        for c in cutlass.range_constexpr(2):
            rmem_st[c] = tCompute[2 * i + c].to(cutlass.BFloat16)
        cute.autovec_copy(rmem_st, cute.slice_(sRht_pairs, ((None, None), (token_base + i, feat_pair))))


@cute.jit
def hadamard_rmem_colwise_fwht_quant(
    rmem_bf16,
    d_buffer,
    tidx,
    norm_const,
    sRht,
    sSf,
    sf_row_base,
    sf_dtype,
    gRht=None,
    seg16=None,
    pitch16=None,
    feat0=None,
    tok16=None,
    row_blocks=None,
    gSf=None,
):
    """Colwise FWHT + NVFP4 quantization from the load_colwise_pairs_bf16 registers,
    stored transposed: packed nibbles pair adjacent tokens of one feature. The
    default target stages into sRht for a transposed TMA store; with gRht set,
    stores go directly into a flat per-expert ragged output buffer."""
    n_vals = 2 * HADAMARD_SIZE
    tCompute = cute.make_rmem_tensor((n_vals,), cutlass.Float32)
    for i in cutlass.range_constexpr(n_vals):
        tCompute[i] = rmem_bf16[i].to(cutlass.Float32)

    _colwise_fwht_inplace(tCompute)

    # Orthonormal scale, rounded through bf16 (== the bf16 output values).
    for i in cutlass.range_constexpr(n_vals):
        tCompute[i] = (tCompute[i] * cutlass.Float32(0.25)).to(cutlass.BFloat16).to(cutlass.Float32)

    _nvfp4_quant_colwise_transposed(
        tCompute,
        d_buffer,
        tidx,
        norm_const,
        sRht,
        sSf,
        sf_row_base,
        sf_dtype,
        gRht,
        seg16,
        pitch16,
        feat0,
        tok16,
        row_blocks,
        gSf,
    )


@cute.jit
def _nvfp4_quant_colwise_transposed(
    tCompute,
    d_buffer,
    tidx,
    norm_const,
    sRht,
    sSf,
    sf_row_base,
    sf_dtype,
    gRht=None,
    seg16=None,
    pitch16=None,
    feat0=None,
    tok16=None,
    row_blocks=None,
    gSf=None,
):
    """Colwise NVFP4 quantization of two 16-token feature columns.

    Data bytes are stored in transposed order, so each feature's 16 transformed
    token values are one contiguous 8-byte vector. ``gRht``/``gSf`` enable direct
    flat per-expert stores for data and per-expert swizzled scale segments.
    """
    token_block = tidx // HADAMARD_SIZE
    feat_pair = tidx % HADAMARD_SIZE
    feature = 2 * feat_pair

    # group_rht_cast's exact (fast_math=0) op sequence — NOT the flashinfer one:
    # gem = ge * (1/6) is pre-folded, and the encode scale is computed with EXACT
    # f32 divisions (enc = 1/(dec * gd), gd = 1/ge). The flashinfer
    # rcp_approx(dec) * ge form agrees at ge = 1 but drifts 1 ulp at non-dyadic
    # ge, flipping e2m1 codes on rounding boundaries.
    gem = norm_const * cutlass.Float32(get_dtype_rcp_limits(cutlass.Float4E2M1FN))
    gd = cutlass.Float32(1.0) / norm_const
    pv = cute.make_rmem_tensor((2,), cutlass.Float32)
    for c in cutlass.range_constexpr(2):
        pv[c] = cutlass.Float32(0.0)
        for i in cutlass.range_constexpr(HADAMARD_SIZE):
            v = tCompute[2 * i + c]
            pv[c] = fmax(pv[c], fmax(v, -v))
    pv[0], pv[1] = cute.arch.mul_packed_f32x2((pv[0], pv[1]), (gem, gem))

    # f32 -> fp8 scale format as a padded 4-wide vector (and widen back the same way):
    # nvgpu.cvt_fptrunc requires a 32-bit-aligned 1-d vector (4 x f8), never a
    # scalar. Same pattern as grouped_gemm_swiglu_quant's pvscale round-trip.
    pv_f32x4 = cute.make_rmem_tensor((4,), cutlass.Float32)
    for c in cutlass.range_constexpr(4):
        pv_f32x4[c] = pv[min(c, 1)]
    tCrSFC_f8x4 = cute.make_rmem_tensor((4,), sf_dtype)
    tCrSFC_f8x4.store(pv_f32x4.load().to(sf_dtype))
    tCrSFC_f32x4 = cute.make_rmem_tensor((4,), cutlass.Float32)
    tCrSFC_f32x4.store(tCrSFC_f8x4.load().to(cutlass.Float32))
    if cutlass.const_expr(gSf is not None):
        gSf = cute.recast_tensor(gSf, sf_dtype)
    for c in cutlass.range_constexpr(2):
        if cutlass.const_expr(gSf is not None):
            row = feat0 + sf_row_base + feature + c
            col = tok16 + token_block
            col_blocks = pitch16 // 4
            sf_idx = (
                seg16
                + (row % 32) * 4 * row_blocks * 4 * col_blocks
                + ((row // 32) % 4) * row_blocks * 4 * col_blocks
                + (row // 128) * 4 * col_blocks
                + (col % 4) * col_blocks
                + (col // 4)
            )
            gSf[sf_idx] = tCrSFC_f8x4[c]
        else:
            sSf[(sf_row_base + feature + c, token_block)] = tCrSFC_f8x4[c]

    fp32_max = cutlass.Float32(3.40282346638528859812e38)
    acc_scale_min0 = fmin(cutlass.Float32(1.0) / (tCrSFC_f32x4[0] * gd), fp32_max, nan=True)
    acc_scale_min1 = fmin(cutlass.Float32(1.0) / (tCrSFC_f32x4[1] * gd), fp32_max, nan=True)
    for i in cutlass.range_constexpr(HADAMARD_SIZE):
        tCompute[2 * i], tCompute[2 * i + 1] = cute.arch.mul_packed_f32x2(
            (tCompute[2 * i], tCompute[2 * i + 1]),
            (acc_scale_min0, acc_scale_min1),
        )

    tCol = cute.make_rmem_tensor((HADAMARD_SIZE,), cutlass.Float32)
    tCol_fp4 = cute.make_rmem_tensor((HADAMARD_SIZE,), cutlass.Float4E2M1FN)
    if cutlass.const_expr(gRht is not None):
        gRht_cols = cute.zipped_divide(gRht, (HADAMARD_SIZE,))
    else:
        sRht_cols = cute.zipped_divide(cute.slice_(sRht, (None, None, d_buffer)), (HADAMARD_SIZE, 1))
    for c in cutlass.range_constexpr(2):
        for i in cutlass.range_constexpr(HADAMARD_SIZE):
            tCol[i] = tCompute[2 * i + c]
        tCol_fp4.store(tCol.load().to(cutlass.Float4E2M1FN))
        if cutlass.const_expr(gRht is not None):
            col = seg16 + (feat0 + sf_row_base + feature + c) * pitch16 + tok16 + token_block
            cute.autovec_copy(tCol_fp4, cute.slice_(gRht_cols, ((None,), col)))
        else:
            cute.autovec_copy(
                tCol_fp4,
                cute.slice_(sRht_cols, ((None, None), (token_block, feature + c))),
            )


@cute.jit
def hadamard_rmem_rowwise_fwht(rmem_bf16, d_buffer, tidx, sRht, norm_const, sSf, sf_pair, sf_dtype):
    """Per-token FWHT over the thread's full 32-feature row (two independent 16-feature
    Hadamard blocks) from the load_row_bf16 registers, stored to sRht at the same
    (token, feature) coords: bf16, or NVFP4 (packed e2m1 into sRht + e4m3/ue5m3 block
    scales into sSf slots (tidx, 2*sf_pair..)) when sRht is fp4-typed."""
    token = tidx
    row_feats = 2 * HADAMARD_SIZE

    tCompute = cute.make_rmem_tensor(rmem_bf16.shape, cutlass.Float32)
    for i in cutlass.range_constexpr(row_feats):
        tCompute[i] = rmem_bf16[i].to(cutlass.Float32)

    for off in cutlass.range_constexpr(0, row_feats, HADAMARD_SIZE):
        for base in cutlass.range_constexpr(0, HADAMARD_SIZE, 2):
            x = tCompute[off + base]
            y = tCompute[off + base + 1]
            tCompute[off + base] = x + y
            tCompute[off + base + 1] = x - y
        for base in cutlass.range_constexpr(0, HADAMARD_SIZE, 4):
            for i in cutlass.range_constexpr(2):
                x = tCompute[off + base + i]
                y = tCompute[off + base + i + 2]
                tCompute[off + base + i] = x + y
                tCompute[off + base + i + 2] = x - y
        for base in cutlass.range_constexpr(0, HADAMARD_SIZE, 8):
            for i in cutlass.range_constexpr(4):
                x = tCompute[off + base + i]
                y = tCompute[off + base + i + 4]
                tCompute[off + base + i] = x + y
                tCompute[off + base + i + 4] = x - y
        for i in cutlass.range_constexpr(8):
            x = tCompute[off + i]
            y = tCompute[off + i + 8]
            tCompute[off + i] = x + y
            tCompute[off + i + 8] = x - y

    # Orthonormal scale, rounded through bf16 (== the bf16 output values) either way.
    for i in cutlass.range_constexpr(row_feats):
        tCompute[i] = (tCompute[i] * cutlass.Float32(0.25)).to(cutlass.BFloat16).to(cutlass.Float32)

    if cutlass.const_expr(sRht.element_type == cutlass.Float4E2M1FN):
        _nvfp4_quant_row(tCompute, d_buffer, token, sRht, norm_const, sSf, sf_pair, sf_dtype)
    else:
        rmem_st = cute.make_rmem_tensor(rmem_bf16.shape, cutlass.BFloat16)
        for i in cutlass.range_constexpr(row_feats):
            rmem_st[i] = tCompute[i].to(cutlass.BFloat16)
        sRht_tiles = cute.zipped_divide(cute.slice_(sRht, (None, None, d_buffer)), (1, row_feats))
        dst = cute.slice_(sRht_tiles, ((None, None), (token, 0)))
        cute.autovec_copy(rmem_st, dst)



# Copyright (c) 2025, Tri Dao, Siyu Wang, Shengbin Di, Yuxi Chi, Johnsonms, Linfeng Zheng, Haoyan Huang, Lanbo Li, Yun Zhong, Man Yuan, Minmin Sun, Yong Li, Wei Lin.
# SPDX-License-Identifier: MIT

from enum import IntEnum, auto
from typing import Optional, Tuple
from dataclasses import dataclass

try:
    from typing import override
except ImportError:  # Python < 3.12
    from typing_extensions import override

import cutlass
from cutlass.pipeline import PipelineClcFetchAsync, PipelineState
from cutlass._mlir import ir
import cutlass.cute as cute
from cutlass import Int32, const_expr
from cutlass.cute import FastDivmodDivisor
from cutlass.utils import ClcDynamicPersistentTileScheduler, ClcDynamicPersistentTileSchedulerParams

from cudnn.block_sparse_attention.csrc.utils.cute_dsl_utils import ParamsBase

from cudnn.block_sparse_attention.csrc.utils import kernel_utils as utils


class SchedulingMode(IntEnum):
    NONE = auto()
    STATIC = auto()
    DYNAMIC = auto()
    CLC = auto()


@dataclass
class ClcState(ParamsBase):
    """Owns the runtime state shared by CLC-capable tile schedulers.

    `BlockSparseAttnForwardSm100Blk128` constructs this state because it owns the CLC
    response buffer, mbarrier storage, and launch geometry needed to initialize
    the hardware scheduler and async pipeline. Individual tile schedulers then
    consume this state and map the returned hardware work tiles into their own
    logical `WorkTileInfo` coordinates.

    To add CLC support to a scheduler:
    - implement `clc_problem_shape(params)` so the kernel can create the hardware scheduler
    - accept `clc: ClcState | None` in `create(...)` / `__init__`
    - map `clc.initial_work_tile_info()` and `clc.get_current_work()` into scheduler coordinates
    """

    _hw_scheduler: ClcDynamicPersistentTileScheduler
    _pipeline: PipelineClcFetchAsync
    _consumer_state: PipelineState
    _producer_state: PipelineState

    @staticmethod
    def create(
        *,
        hw_scheduler: ClcDynamicPersistentTileScheduler,
        pipeline: PipelineClcFetchAsync,
        consumer_state: PipelineState,
        producer_state: PipelineState,
    ) -> "ClcState":
        return ClcState(hw_scheduler, pipeline, consumer_state, producer_state)

    def initial_work_tile_info(self):
        return self._hw_scheduler.initial_work_tile_info()

    def get_current_work(self):
        return self._hw_scheduler.get_current_work()

    def prefetch_next_work(self, *, loc=None, ip=None):
        self._pipeline.producer_acquire(self._producer_state, loc=loc, ip=ip)
        mbarrier_addr = self._pipeline.producer_get_barrier(self._producer_state, loc=loc, ip=ip)
        self._hw_scheduler.advance_to_next_work(mbarrier_addr, loc=loc, ip=ip)
        self._producer_state.advance(loc=loc, ip=ip)

    def consumer_wait(self, *, loc=None, ip=None):
        self._pipeline.consumer_wait(self._consumer_state, loc=loc, ip=ip)

    def consumer_release(self, *, loc=None, ip=None):
        self._pipeline.consumer_release(self._consumer_state, loc=loc, ip=ip)
        self._consumer_state.advance(loc=loc, ip=ip)

    def producer_tail(self, *, loc=None, ip=None):
        self._pipeline.producer_tail(self._producer_state, loc=loc, ip=ip)


class WorkTileInfo(cutlass.utils.WorkTileInfo):
    """Altered WorkTileInfo which includes four axes: (block, head, batch, split)"""

    @override
    def __new_from_mlir_values__(self, values: list[ir.Value]) -> "WorkTileInfo":
        assert len(values) == 5
        new_tile_idx = cutlass.new_from_mlir_values(self._tile_idx, values[:-1])
        new_is_valid_tile = cutlass.new_from_mlir_values(self._is_valid_tile, [values[-1]])
        return WorkTileInfo(new_tile_idx, new_is_valid_tile)


@dataclass
class TileSchedulerArguments(ParamsBase):
    num_block: Int32
    num_head: Int32
    num_batch: Int32
    num_splits: Int32
    seqlen_k: Int32
    headdim: Int32
    headdim_v: Int32
    total_q: Int32
    tile_shape_mn: cutlass.Constexpr[Tuple[int, int]]
    mCuSeqlensQ: Optional[cute.Tensor] = None
    mSeqUsedQ: Optional[cute.Tensor] = None
    qhead_per_kvhead_packgqa: cutlass.Constexpr[int] = 1
    element_size: cutlass.Constexpr[int] = 2
    is_persistent: cutlass.Constexpr[bool] = False
    lpt: cutlass.Constexpr[bool] = False
    is_split_kv: cutlass.Constexpr[bool] = False
    head_swizzle: cutlass.Constexpr[bool] = False


class SingleTileScheduler:
    @dataclass
    class Params(ParamsBase):
        num_block: Int32
        num_head: Int32
        num_batch: Int32
        num_splits: Int32
        num_splits_divmod: FastDivmodDivisor
        is_split_kv: cutlass.Constexpr[bool] = False

        @staticmethod
        def create(args: TileSchedulerArguments, *, loc=None, ip=None) -> "SingleTileScheduler.Params":
            return SingleTileScheduler.Params(
                args.num_block,
                args.num_head,
                args.num_batch,
                args.num_splits,
                FastDivmodDivisor(args.num_splits),
                args.is_split_kv,
            )

    def __init__(self, params: Params, blk_coord: cute.Coord, *, loc=None, ip=None):
        self.params = params
        self._blk_coord = blk_coord
        self._is_first_block = True
        self._loc = loc
        self._ip = ip

    @staticmethod
    def to_underlying_arguments(
        args: TileSchedulerArguments,
        *,
        scheduling_mode: SchedulingMode = SchedulingMode.STATIC,
        loc=None,
        ip=None,
    ) -> Params:
        assert scheduling_mode == SchedulingMode.STATIC, f"SingleTileScheduler only supports STATIC, got {scheduling_mode!r}"
        return SingleTileScheduler.Params.create(args, loc=loc, ip=ip)

    @staticmethod
    def create(params: Params, clc: ClcState | None = None, *, loc=None, ip=None) -> "SingleTileScheduler":
        return SingleTileScheduler(params, cute.arch.block_idx(), loc=loc, ip=ip)

    # called by host
    @staticmethod
    def get_grid_shape(
        params: Params,
        *,
        loc=None,
        ip=None,
    ) -> Tuple[Int32, Int32, Int32]:
        return (
            params.num_block,
            params.num_head * params.num_splits,
            params.num_batch,
        )

    def get_current_work(self, *, loc=None, ip=None) -> WorkTileInfo:
        block_idx, head_idx, batch_idx = self._blk_coord
        if const_expr(self.params.is_split_kv):
            head_idx, split_idx = divmod(head_idx, self.params.num_splits_divmod)
        else:
            split_idx = Int32(0)
        return WorkTileInfo(
            (block_idx, head_idx, batch_idx, split_idx),
            self._is_first_block,
        )

    def initial_work_tile_info(self, *, loc=None, ip=None):
        return self.get_current_work(loc=loc, ip=ip)

    def prefetch_next_work(self, *, loc=None, ip=None):
        pass

    def advance_to_next_work(self, *, loc=None, ip=None):
        self._is_first_block = False
        return self.get_current_work()

    def consumer_advance(self, *, loc=None, ip=None):
        return self.advance_to_next_work(loc=loc, ip=ip)

    def producer_tail(self, *, loc=None, ip=None):
        pass

    def __extract_mlir_values__(self):
        values, self._values_pos = [], []
        for obj in [self.params, self._blk_coord]:
            obj_values = cutlass.extract_mlir_values(obj)
            values += obj_values
            self._values_pos.append(len(obj_values))
        return values

    def __new_from_mlir_values__(self, values):
        obj_list = []
        for obj, n_items in zip([self.params, self._blk_coord], self._values_pos):
            obj_list.append(cutlass.new_from_mlir_values(obj, values[:n_items]))
            values = values[n_items:]
        return SingleTileScheduler(*(tuple(obj_list)), loc=self._loc)


class SingleTileVarlenScheduler:
    @dataclass
    class Params(ParamsBase):
        num_head: Int32
        num_batch: Int32
        total_q: Int32
        num_splits: Int32
        max_kvblock_in_l2: Int32
        tile_shape_mn: cutlass.Constexpr[Tuple[int, int]]
        mCuSeqlensQ: Optional[cute.Tensor] = None
        mSeqUsedQ: Optional[cute.Tensor] = None
        qhead_per_kvhead_packgqa: cutlass.Constexpr[int] = 1
        lpt: cutlass.Constexpr[bool] = False
        is_split_kv: cutlass.Constexpr[bool] = False
        head_swizzle: cutlass.Constexpr[bool] = False
        scheduling_mode: cutlass.Constexpr[SchedulingMode] = SchedulingMode.STATIC

        @staticmethod
        @cute.jit
        def create(
            args: TileSchedulerArguments,
            *,
            scheduling_mode: SchedulingMode = SchedulingMode.STATIC,
            loc=None,
            ip=None,
        ) -> "SingleTileVarlenScheduler.Params":
            assert scheduling_mode in (SchedulingMode.STATIC, SchedulingMode.CLC), f"Only STATIC and CLC are supported, got {scheduling_mode!r}"
            size_l2 = 50 * 1024 * 1024  # 50 MB for K & V
            # if backward, this is qdo block size
            kv_block_size = (args.headdim + args.headdim_v) * args.element_size * args.tile_shape_mn[1]
            # if backward, add dqaccum block size to calculate swizzle
            if args.head_swizzle:
                kv_block_size += args.headdim * 4 * args.tile_shape_mn[1]
            max_kvblock_in_l2 = size_l2 // kv_block_size
            assert args.mCuSeqlensQ is not None or args.mSeqUsedQ is not None, "At least one of mCuSeqlensQ or mSeqUsedQ must be provided"
            return SingleTileVarlenScheduler.Params(
                num_head=args.num_head,
                num_batch=args.num_batch,
                total_q=args.total_q,
                num_splits=args.num_splits,
                max_kvblock_in_l2=max_kvblock_in_l2,
                tile_shape_mn=args.tile_shape_mn,
                mCuSeqlensQ=args.mCuSeqlensQ,
                mSeqUsedQ=args.mSeqUsedQ,
                qhead_per_kvhead_packgqa=args.qhead_per_kvhead_packgqa,
                lpt=args.lpt,
                is_split_kv=args.is_split_kv,
                head_swizzle=args.head_swizzle,
                scheduling_mode=scheduling_mode,
            )

    def __init__(
        self,
        params: Params,
        tile_idx: Int32,
        split_idx: Int32,
        clc: ClcState | None = None,
        *,
        loc=None,
        ip=None,
    ):
        self.params = params
        self._tile_idx = tile_idx
        self._split_idx = split_idx
        self._is_first_block = True
        self.clc = clc
        self._loc = loc
        self._ip = ip

    @staticmethod
    def to_underlying_arguments(
        args: TileSchedulerArguments,
        *,
        scheduling_mode: SchedulingMode = SchedulingMode.STATIC,
        loc=None,
        ip=None,
    ) -> Params:
        return SingleTileVarlenScheduler.Params.create(args, scheduling_mode=scheduling_mode, loc=loc, ip=ip)

    @staticmethod
    @cute.jit
    def clc_problem_shape(params: Params):
        return ClcDynamicPersistentTileSchedulerParams(
            problem_shape_ntile_mnl=SingleTileVarlenScheduler.get_grid_shape(params),
            cluster_shape_mnk=(1, 1, 1),
        )

    @staticmethod
    @cute.jit
    def create(params: Params, clc: ClcState | None = None, *, loc=None, ip=None) -> "SingleTileVarlenScheduler":
        if const_expr(params.scheduling_mode == SchedulingMode.CLC):
            block_idx = cute.arch.block_idx()
            split_idx = Int32(0)
            if const_expr(params.is_split_kv):
                split_idx = block_idx[1]
            return SingleTileVarlenScheduler(
                params,
                block_idx[0],
                split_idx,
                clc,
                loc=loc,
                ip=ip,
            )
        tile_idx, split_idx, _ = cute.arch.block_idx()
        return SingleTileVarlenScheduler(params, tile_idx, split_idx, loc=loc, ip=ip)

    # called by host
    @staticmethod
    def get_grid_shape(
        params: Params,
        *,
        loc=None,
        ip=None,
    ) -> Tuple[Int32, Int32, Int32]:
        total_blocks_max = (params.total_q + params.num_batch * (params.tile_shape_mn[0] - 1)) // params.tile_shape_mn[0]
        return (total_blocks_max * params.num_head, params.num_splits, Int32(1))

    @cute.jit
    def _get_num_m_blocks(self, lane: Int32, bidb_start: Int32) -> Int32:
        params = self.params
        batch_idx = lane + bidb_start
        if cutlass.const_expr(params.mSeqUsedQ is not None):
            seqlen = Int32(0)
            if batch_idx < params.num_batch:
                seqlen = params.mSeqUsedQ[batch_idx]
        else:
            assert params.mCuSeqlensQ is not None
            cur_cu_seqlen = Int32(0)
            if batch_idx <= params.num_batch:
                cur_cu_seqlen = params.mCuSeqlensQ[batch_idx]
            next_cu_seqlen = cute.arch.shuffle_sync_down(cur_cu_seqlen, offset=1)
            seqlen = next_cu_seqlen - cur_cu_seqlen
        if cutlass.const_expr(params.qhead_per_kvhead_packgqa > 1):
            seqlen *= params.qhead_per_kvhead_packgqa
        return cute.ceil_div(seqlen, params.tile_shape_mn[0]) if batch_idx < params.num_batch and lane < cute.arch.WARP_SIZE - 1 else Int32(0)

    @cute.jit
    def _varlen_coord_map(self) -> WorkTileInfo:
        """Map self._tile_idx to (block, head, batch) via warp-level prefix sums."""
        params = self.params
        lane_idx = cute.arch.lane_idx()
        num_m_blocks = self._get_num_m_blocks(lane_idx, bidb_start=0)
        num_m_blocks_cumulative = utils.warp_prefix_sum(num_m_blocks, lane_idx)
        # Total number of blocks for the next 31 batches
        m_blocks_in_group = cute.arch.shuffle_sync(num_m_blocks_cumulative, cute.arch.WARP_SIZE - 1)
        # Same for all lanes
        group_end_tile = m_blocks_in_group * params.num_head
        block, head_idx, batch_idx = Int32(0), Int32(0), Int32(0)
        next_tile_idx = self._tile_idx
        while group_end_tile <= next_tile_idx:
            batch_idx += cute.arch.WARP_SIZE - 1
            if batch_idx >= params.num_batch:
                batch_idx = Int32(params.num_batch)
                group_end_tile = next_tile_idx + 1
            else:
                num_m_blocks = self._get_num_m_blocks(lane_idx, bidb_start=batch_idx)
                num_m_blocks_cumulative = utils.warp_prefix_sum(num_m_blocks, lane_idx)
                m_blocks_in_group = cute.arch.shuffle_sync(num_m_blocks_cumulative, cute.arch.WARP_SIZE - 1)
                group_end_tile += m_blocks_in_group * params.num_head
        is_valid = False
        if batch_idx >= params.num_batch:
            block, head_idx, batch_idx = Int32(0), Int32(0), Int32(params.num_batch)
        else:
            group_start_tile = group_end_tile - m_blocks_in_group * params.num_head
            # The next problem to process is the first one that does not have ending tile position
            # that is greater than or equal to tile index.
            batch_idx_in_group = cute.arch.popc(cute.arch.vote_ballot_sync(group_start_tile + num_m_blocks_cumulative * params.num_head <= next_tile_idx))
            batch_idx += batch_idx_in_group
            num_m_blocks_prev_lane = 0 if batch_idx_in_group == 0 else cute.arch.shuffle_sync(num_m_blocks_cumulative, batch_idx_in_group - 1)
            num_m_blocks = cute.arch.shuffle_sync(num_m_blocks, batch_idx_in_group)
            mh_block = next_tile_idx - group_start_tile - num_m_blocks_prev_lane * params.num_head
            if cutlass.const_expr(params.lpt or params.head_swizzle):
                # Apply longest-processing-time ordering while accounting for
                # sequence lengths that vary by batch.
                # TODO: is there any case where num_m_blocks is 0?
                # TODO: by right we should read the seqlen_kv but we're assuming seqlen_q == seqlen_k here
                num_n_blocks = num_m_blocks * params.tile_shape_mn[0] // params.qhead_per_kvhead_packgqa // params.tile_shape_mn[1]
                # Seems faster to have this be a power of 2
                nheads_in_l2 = (
                    16
                    if num_n_blocks * 16 <= params.max_kvblock_in_l2
                    else (
                        8
                        if num_n_blocks * 8 <= params.max_kvblock_in_l2
                        else (4 if num_n_blocks * 4 <= params.max_kvblock_in_l2 else (2 if num_n_blocks * 2 <= params.max_kvblock_in_l2 else 1))
                    )
                )
                nheads_in_l2 = min(nheads_in_l2, params.num_head)
                mh_in_l2 = nheads_in_l2 * num_m_blocks
                section_idx = mh_block // mh_in_l2
                l2_mod = mh_block - section_idx * mh_in_l2
                # Deal with tail section
                nheads_in_this_section = nheads_in_l2 if nheads_in_l2 * (section_idx + 1) <= params.num_head else params.num_head - section_idx * nheads_in_l2
                block = l2_mod // nheads_in_this_section
                head_idx_residual = l2_mod - block * nheads_in_this_section
                head_idx = section_idx * nheads_in_l2 + head_idx_residual
                if cutlass.const_expr(params.lpt):
                    block = num_m_blocks - 1 - block
            else:
                head_idx = mh_block // num_m_blocks
                block = mh_block - head_idx * num_m_blocks
            is_valid = self._is_first_block and batch_idx < params.num_batch
        split_idx = self._split_idx if const_expr(params.is_split_kv) else Int32(0)
        return WorkTileInfo((Int32(block), Int32(head_idx), Int32(batch_idx), split_idx), is_valid)

    @cute.jit
    def get_current_work(self, *, loc=None, ip=None) -> WorkTileInfo:
        if const_expr(self.params.scheduling_mode == SchedulingMode.CLC):
            clc_work = self.clc.get_current_work()
            # Default to grid_dim (one past last valid flat index) so _varlen_coord_map
            # returns is_valid=False when CLC is exhausted. CLC tile_idx is garbage when
            # invalid, so we can't trust it. Local-then-assign avoids CuTe DSL structural
            # mismatch on self inside the runtime if.
            new_tile_idx = cute.arch.grid_dim()[0]
            new_split_idx = Int32(0)
            if clc_work.is_valid_tile:
                new_tile_idx = clc_work.tile_idx[0]
                if const_expr(self.params.is_split_kv):
                    new_split_idx = clc_work.tile_idx[1]
            self._tile_idx = new_tile_idx
            self._split_idx = new_split_idx
        return self._varlen_coord_map()

    @cute.jit
    def initial_work_tile_info(self, *, loc=None, ip=None):
        if const_expr(self.params.scheduling_mode == SchedulingMode.CLC):
            clc_work = self.clc.initial_work_tile_info()
            # See get_current_work for why grid_dim and local-then-assign.
            new_tile_idx = cute.arch.grid_dim()[0]
            new_split_idx = Int32(0)
            if clc_work.is_valid_tile:
                new_tile_idx = clc_work.tile_idx[0]
                if const_expr(self.params.is_split_kv):
                    new_split_idx = clc_work.tile_idx[1]
            self._tile_idx = new_tile_idx
            self._split_idx = new_split_idx
        return self._varlen_coord_map()

    def prefetch_next_work(self, *, loc=None, ip=None):
        if const_expr(self.params.scheduling_mode == SchedulingMode.CLC):
            self.clc.prefetch_next_work(loc=loc, ip=ip)

    def advance_to_next_work(self, *, loc=None, ip=None):
        if const_expr(self.params.scheduling_mode == SchedulingMode.CLC):
            self.clc.consumer_wait(loc=loc, ip=ip)
            work = self.get_current_work()
            self.clc.consumer_release(loc=loc, ip=ip)
            return work
        self._is_first_block = False
        return self.get_current_work()

    def producer_tail(self, *, loc=None, ip=None):
        if const_expr(self.params.scheduling_mode == SchedulingMode.CLC):
            self.clc.producer_tail(loc=loc, ip=ip)

    def __extract_mlir_values__(self):
        values, self._values_pos = [], []
        objs = [self.params, self._tile_idx, self._split_idx]
        if const_expr(self.params.scheduling_mode == SchedulingMode.CLC):
            objs += [self.clc]
        for obj in objs:
            obj_values = cutlass.extract_mlir_values(obj)
            values += obj_values
            self._values_pos.append(len(obj_values))
        return values

    def __new_from_mlir_values__(self, values):
        obj_list = []
        objs = [self.params, self._tile_idx, self._split_idx]
        if const_expr(self.params.scheduling_mode == SchedulingMode.CLC):
            objs += [self.clc]
        for obj, n_items in zip(objs, self._values_pos):
            obj_list.append(cutlass.new_from_mlir_values(obj, values[:n_items]))
            values = values[n_items:]
        return self.__class__(*obj_list, loc=self._loc)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from .agg_simple import agg_simple
from .graph import CscGraph

__all__ = [
    "CscGraph",
    "agg_simple",
]



# Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT


from cutlass.experimental import primitives as nvvm
import cutlass
import cutlass.cute as cute


@cute.jit
def tmem_alloc(tmem_ptr_i32, num_cols: int, cta_group_kind, is_exclusive: cutlass.Constexpr[bool] = False):
    # is_exclusive is the sm_107/sm_109 >512-column mechanism (GR100's 576-col
    # TMEM; cute.arch.tmem.is_tmem_allocation_exclusive) — pass True only
    # there.  It is fenced out of the PUBLIC cutlass-dsl tcgen05_alloc
    # wrapper, so the kwarg is only emitted on the True branch
    # (internal-wheel-only callers); <=512-column allocations never trace it.
    if cutlass.const_expr(is_exclusive):
        nvvm.tcgen05_alloc(
            tmem_ptr_i32,
            cutlass.Int32(num_cols),
            is_exclusive=True,
            group=cta_group_kind,
        )
    else:
        nvvm.tcgen05_alloc(
            tmem_ptr_i32,
            cutlass.Int32(num_cols),
            group=cta_group_kind,
        )
    nvvm.tcgen05_relinquish_alloc_permit(group=cta_group_kind)
    nvvm.bar_warp_sync(cute.arch.FULL_MASK)


@cute.jit
def tmem_dealloc(tmem_ptr_i32, num_cols: int, cta_group_kind):
    tmem_ptr_for_dealloc = nvvm.make_tmem_ptr(tmem_ptr_i32.load(), cutlass.Int8)
    nvvm.tcgen05_dealloc(
        tmem_ptr_for_dealloc,
        cutlass.Int32(num_cols),
        group=cta_group_kind,
    )



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""cuDNN-accelerated drop-in for FLA 0.5.2's plain ``GatedMLP``.

The adapter is deliberately narrow.  It replaces only the arithmetic of the
stock, local, bias-free BF16 ``swish`` module with
``cudnn.gemm.ops.swiglu_mlp``.  Anything whose semantics may differ -- tensor
parallelism, tensor/module subclasses, quantization, LoRA, parametrizations,
hooks, non-compact storage, another FLA version, or graph compilation -- calls
the original FLA ``forward`` unchanged.
"""

from __future__ import annotations

import functools
from importlib import metadata

import torch

import cudnn

_SUPPORTED_FLA_VERSION = "0.5.2"
_DECLINE_ERRORS = (NotImplementedError, cudnn.cudnnGraphNotSupportedError, ImportError)
_LAST = {"path": None}

_MODULE_HOOK_ATTRS = (
    "_backward_hooks",
    "_backward_pre_hooks",
    "_forward_hooks",
    "_forward_pre_hooks",
)
_GLOBAL_MODULE_HOOK_ATTRS = (
    "_global_backward_hooks",
    "_global_backward_pre_hooks",
    "_global_forward_hooks",
    "_global_forward_pre_hooks",
)
_TENSOR_HOOK_ATTRS = ("_backward_hooks", "_post_accumulate_grad_hooks")


def last_path() -> str | None:
    """The route the most recent shimmed MLP call took, for telemetry/tests."""
    return _LAST["path"]


def _installed_fla_version() -> str | None:
    try:
        return metadata.version("flash-linear-attention")
    except metadata.PackageNotFoundError:
        return None


def _supports_installed_fla() -> bool:
    return _installed_fla_version() == _SUPPORTED_FLA_VERSION


def _is_compiling() -> bool:
    compiler = getattr(torch, "compiler", None)
    if compiler is not None and compiler.is_compiling():
        return True
    dynamo = getattr(torch, "_dynamo", None)
    return bool(dynamo is not None and dynamo.is_compiling())


def _is_cuda_autocast_enabled() -> bool:
    try:
        return torch.is_autocast_enabled("cuda")
    except TypeError:  # torch versions whose no-argument form means CUDA
        return torch.is_autocast_enabled()


def _cuda_autocast_dtype():
    if not _is_cuda_autocast_enabled():
        return None
    try:
        return torch.get_autocast_dtype("cuda")
    except (AttributeError, TypeError):
        return torch.get_autocast_gpu_dtype()


def _is_cuda_tensor(tensor) -> bool:
    return tensor.is_cuda


def _device_capability(device) -> tuple[int, int]:
    return torch.cuda.get_device_capability(device)


def _call_native(x, gate_weight, up_weight, down_weight):
    # Keep importing cudnn.fla independent of the optional GEMM/CuTeDSL stack;
    # an unavailable native op is a typed decline to the original FLA method.
    from cudnn.gemm.ops import swiglu_mlp

    return swiglu_mlp(x, gate_weight, up_weight, down_weight)


def _has_hooks(obj, attrs) -> bool:
    return any(bool(getattr(obj, attr, None)) for attr in attrs)


def _has_global_module_hooks() -> bool:
    module_impl = torch.nn.modules.module
    return _has_hooks(module_impl, _GLOBAL_MODULE_HOOK_ATTRS)


def _is_parametrized(module) -> bool:
    parametrize = getattr(torch.nn.utils, "parametrize", None)
    return bool(parametrize is not None and parametrize.is_parametrized(module))


def _plain_parameter_decline(parameter, *, shape, device) -> str | None:
    if type(parameter) is not torch.nn.Parameter:
        return "tensor-subclass"
    if type(parameter.data) is not torch.Tensor or parameter.layout is not torch.strided:
        return "tensor-subclass"
    if parameter.dtype is not torch.bfloat16:
        return "non-bf16"
    if not _is_cuda_tensor(parameter) or parameter.device != device:
        return "nonlocal-device"
    if tuple(parameter.shape) != tuple(shape):
        return "shape"
    if not parameter.is_contiguous():
        return "noncontiguous"
    if not parameter.is_leaf or parameter.grad_fn is not None:
        return "nonlocal-parameter"
    if _has_hooks(parameter, _TENSOR_HOOK_ATTRS):
        return "hooks"
    return None


def _module_decline(module, *, expected_type, expected_parameter_names) -> str | None:
    if _is_parametrized(module):
        return "parametrized"
    if type(module) is not expected_type:
        return "custom-module"
    if "forward" in module.__dict__:
        return "custom-forward"
    if set(module._parameters) != set(expected_parameter_names) or module._buffers or module._modules:
        return "custom-module"
    if _has_hooks(module, _MODULE_HOOK_ATTRS):
        return "hooks"
    return None


def _decline_reason(self, x, *, gated_mlp_cls, swiglu_linear_cls, fla_version) -> str | None:
    # This check must precede all tensor/device introspection so torch.compile
    # traces the incumbent implementation rather than specializing this adapter.
    if _is_compiling() or torch.jit.is_scripting() or torch.jit.is_tracing():
        return "compile"
    if _cuda_autocast_dtype() not in (None, torch.bfloat16):
        # Stock FLA runs all three Linear modules under autocast.  Entering the
        # BF16-only fused op would instead bypass a different requested dtype.
        # BF16 autocast preserves the admitted operands/results and remains fast.
        return "autocast"
    if fla_version != _SUPPORTED_FLA_VERSION:
        return "fla-version"
    if type(self) is not gated_mlp_cls or "forward" in self.__dict__:
        return "custom-gated-mlp"
    if self._parameters or self._buffers or set(self._modules) != {"gate_proj", "up_proj", "down_proj", "swiglu_linear"}:
        return "custom-gated-mlp"
    if _is_parametrized(self) or _has_hooks(self, _MODULE_HOOK_ATTRS) or _has_global_module_hooks():
        return "hooks-or-parametrization"
    if getattr(self, "hidden_act", None) != "swish" or getattr(self, "fuse_swiglu", None) is not True:
        return "variant"

    projections = (self.gate_proj, self.up_proj, self.down_proj)
    for projection in projections:
        reason = _module_decline(projection, expected_type=torch.nn.Linear, expected_parameter_names=("weight", "bias"))
        if reason is not None:
            return reason
        if projection.bias is not None:
            return "bias"

    reason = _module_decline(self.swiglu_linear, expected_type=swiglu_linear_cls, expected_parameter_names=())
    if reason is not None:
        return reason

    if type(x) is not torch.Tensor or x.layout is not torch.strided:
        return "tensor-subclass"
    if x.dtype is not torch.bfloat16:
        return "non-bf16"
    if not _is_cuda_tensor(x):
        return "non-cuda"
    if x.dim() < 2 or x.numel() == 0 or not x.is_contiguous():
        return "shape-or-layout"
    if _has_hooks(x, _TENSOR_HOOK_ATTRS):
        return "hooks"
    if _device_capability(x.device)[0] != 10:
        return "non-sm100"

    hidden = getattr(self, "hidden_size", None)
    intermediate = getattr(self, "intermediate_size", None)
    if not isinstance(hidden, int) or not isinstance(intermediate, int) or hidden <= 0 or intermediate <= 0:
        return "shape"
    if hidden % 8 or intermediate % 8 or x.shape[-1] != hidden:
        return "shape"
    expected_linears = (
        (self.gate_proj, hidden, intermediate),
        (self.up_proj, hidden, intermediate),
        (self.down_proj, intermediate, hidden),
    )
    for projection, in_features, out_features in expected_linears:
        if projection.in_features != in_features or projection.out_features != out_features:
            return "shape"
        reason = _plain_parameter_decline(projection.weight, shape=(out_features, in_features), device=x.device)
        if reason is not None:
            return reason
    return None


def make_gated_mlp_forward(real_forward, gated_mlp_cls, swiglu_linear_cls):
    """Wrap FLA's class method with a fail-closed cuDNN SwiGLU-MLP path.

    The wrapper patches the class, so it covers both existing and future
    instances.  FLA 0.5.2 accepts arbitrary ``**kwargs`` and ignores them; the
    native path deliberately does the same, while fallback forwards the exact
    mapping to ``real_forward``.
    """

    fla_version = _installed_fla_version()

    @functools.wraps(real_forward)
    def forward(self, x, **kwargs):
        reason = _decline_reason(
            self,
            x,
            gated_mlp_cls=gated_mlp_cls,
            swiglu_linear_cls=swiglu_linear_cls,
            fla_version=fla_version,
        )
        if reason is not None:
            _LAST["path"] = f"fallback:{reason}"
            return real_forward(self, x, **kwargs)
        # Mark the selected route before entering the custom autograd op.
        # Non-reentrant checkpointing deliberately raises an internal
        # control-flow exception while replaying the op; it must pass through
        # untouched without being misreported as a native execution failure.
        _LAST["path"] = "native"
        try:
            out = _call_native(x, self.gate_proj.weight, self.up_proj.weight, self.down_proj.weight)
        except _DECLINE_ERRORS as error:
            _LAST["path"] = f"fallback:{type(error).__name__}"
            return real_forward(self, x, **kwargs)
        except (RuntimeError, ValueError) as error:
            _LAST["path"] = f"error:{type(error).__name__}"
            raise
        return out

    forward.__cudnn_fla_target__ = "gated_mlp"
    return forward


__all__ = ["last_path", "make_gated_mlp_forward"]



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""Execution engines for pygraph.

Pluggable engines in one flat engine-id space with the cuDNN backend. At
``create_execution_plans()`` the heuristics rank the engines that claim the graph
against the backend's own recommendation into ONE list (``graph.plans``);
``build_plans()`` walks it. Graph construction stays engine-agnostic.

Which engines exist is the library's own business: ``manifest.py`` is the static
table this version ships, matched against a graph by a cheap node-type key
before any engine module is imported. It is the ONLY way a python engine
exists -- there is no registration call, so an engine id always decodes back to
a family and a slot.
"""

from .base import BaseEngine, decline_types, CompiledPlan, ExecutionContext, PlanConfig
from .engine_ids import (
    BACKEND_ENGINE_ID_BASE,
    BACKEND_HEURISTIC_ENGINE_ID,
    CPP_OSS_ENGINE_ID_BASE,
    PYTHON_ENGINE_ID_BASE,
    is_backend_engine,
    is_python_engine,
)
from .manifest import MANIFEST, EngineFamily


def __getattr__(name: str):
    # Engine classes stay reachable by name (tests, advanced callers) but are
    # imported on demand: an engine module drags in torch / the CuTe DSL and
    # `import cudnn` must not pay that. The manifest is the source of truth for
    # what exists; this only resolves a class name to its row.
    for row in MANIFEST:
        if row.factory == name:
            import importlib

            cls = getattr(importlib.import_module(row.module), name)
            globals()[name] = cls
            return cls
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__ = [
    "BaseEngine",
    "CompiledPlan",
    "ExecutionContext",
    "PlanConfig",
    "decline_types",
    "MANIFEST",
    "EngineFamily",
    "BACKEND_ENGINE_ID_BASE",
    "CPP_OSS_ENGINE_ID_BASE",
    "PYTHON_ENGINE_ID_BASE",
    "BACKEND_HEURISTIC_ENGINE_ID",
    "is_python_engine",
    "is_backend_engine",
]



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""Python API and Torch custom-op implementation for GNN simple aggregation."""

from typing import Dict, Optional, Tuple

import torch
from torch import Tensor

from ._dtypes import TORCH_DTYPE_TO_CUDNN, TORCH_INDEX_DTYPE_TO_CUDNN
from ._utils import require_backend_symbols, tensor_pointer, validate_csc_graph
from .graph import CscGraph

# Torch custom-op implementation

_AGGREGATION_TO_INT: Dict[str, int] = {
    "sum": 0,
    "mean": 1,
    "max": 2,
    "min": 3,
}


def _validate_aggregation(aggr: str) -> None:
    if aggr not in _AGGREGATION_TO_INT:
        supported = ", ".join(_AGGREGATION_TO_INT)
        raise ValueError(f"Unsupported aggregation '{aggr}'. Supported: {supported}.")


def _validate_graph(
    offsets: Tensor,
    indices: Tensor,
    map_csc_to_coo: Optional[Tensor],
    num_src_nodes: int,
) -> Tuple[int, int]:
    num_dst_nodes, num_edges = validate_csc_graph(offsets, indices, map_csc_to_coo, num_src_nodes)
    if num_dst_nodes > 0 and num_edges == 0:
        raise ValueError("cuDNN AggSimple does not yet support a nonempty destination set with zero edges")
    return num_dst_nodes, num_edges


def _validate_features(
    offsets: Tensor,
    num_src_nodes: int,
    num_dst_nodes: int,
    num_edges: int,
    node_features: Optional[Tensor],
    edge_features: Optional[Tensor],
    concat_features: Optional[Tensor],
) -> Tuple[torch.dtype, int, int, int]:
    if node_features is None and edge_features is None:
        raise ValueError("at least one of node_features or edge_features must be provided")

    feature_tensors = [tensor for tensor in (node_features, edge_features, concat_features) if tensor is not None]
    dtype = feature_tensors[0].dtype
    if dtype not in TORCH_DTYPE_TO_CUDNN:
        raise TypeError(f"feature tensors must have dtype float32, float16, or bfloat16, got {dtype}")

    for tensor in feature_tensors:
        if tensor.ndim != 2:
            raise ValueError(f"feature tensors must be rank 2, got shape {tuple(tensor.shape)}")
        if not tensor.is_cuda:
            raise ValueError(f"feature tensors must be CUDA tensors, got {tensor.device}")
        if tensor.device != offsets.device:
            raise ValueError(f"all graph and feature tensors must be on {offsets.device}, got {tensor.device}")
        if tensor.dtype != dtype:
            raise TypeError(f"all feature tensors must have dtype {dtype}, got {tensor.dtype}")

    node_feat_dim = 0
    if node_features is not None:
        if node_features.shape[0] != num_src_nodes:
            raise ValueError(f"node_features must have {num_src_nodes} rows, got {node_features.shape[0]}")
        node_feat_dim = node_features.shape[1]

    edge_feat_dim = 0
    if edge_features is not None:
        if edge_features.shape[0] != num_edges:
            raise ValueError(f"edge_features must have {num_edges} rows, got {edge_features.shape[0]}")
        edge_feat_dim = edge_features.shape[1]

    concat_feat_dim = 0
    if concat_features is not None:
        if concat_features.shape[0] != num_dst_nodes:
            raise ValueError(f"concat_features must have {num_dst_nodes} rows, got {concat_features.shape[0]}")
        concat_feat_dim = concat_features.shape[1]

    if node_feat_dim == 0 and edge_feat_dim == 0:
        raise ValueError("node_features and edge_features cannot both have zero feature dimension")

    return dtype, node_feat_dim, edge_feat_dim, concat_feat_dim


def _validate_inputs(
    offsets: Tensor,
    indices: Tensor,
    map_csc_to_coo: Optional[Tensor],
    node_features: Optional[Tensor],
    edge_features: Optional[Tensor],
    concat_features: Optional[Tensor],
    num_src_nodes: int,
    aggr: str,
) -> Tuple[int, int, torch.dtype, int, int, int]:
    _validate_aggregation(aggr)
    num_dst_nodes, num_edges = _validate_graph(offsets, indices, map_csc_to_coo, num_src_nodes)
    dtype, node_feat_dim, edge_feat_dim, concat_feat_dim = _validate_features(
        offsets, num_src_nodes, num_dst_nodes, num_edges, node_features, edge_features, concat_features
    )
    return num_dst_nodes, num_edges, dtype, node_feat_dim, edge_feat_dim, concat_feat_dim


def _require_backend() -> None:
    require_backend_symbols("gnn_agg_simple_forward", "gnn_agg_simple_backward")


def forward(
    offsets: Tensor,
    indices: Tensor,
    map_csc_to_coo: Optional[Tensor],
    node_features: Optional[Tensor],
    edge_features: Optional[Tensor],
    concat_features: Optional[Tensor],
    num_src_nodes: int,
    aggr: str,
) -> Tuple[Tensor, Tensor]:
    num_dst_nodes, num_edges, dtype, node_feat_dim, edge_feat_dim, concat_feat_dim = _validate_inputs(
        offsets, indices, map_csc_to_coo, node_features, edge_features, concat_features, num_src_nodes, aggr
    )

    offsets = offsets.contiguous()
    indices = indices.contiguous()
    map_csc_to_coo = None if map_csc_to_coo is None else map_csc_to_coo.contiguous()
    node_features = None if node_features is None else node_features.contiguous()
    edge_features = None if edge_features is None else edge_features.contiguous()
    concat_features = None if concat_features is None else concat_features.contiguous()

    output = torch.empty((num_dst_nodes, node_feat_dim + edge_feat_dim + concat_feat_dim), device=offsets.device, dtype=dtype)
    if aggr in ("max", "min"):
        out_positions = torch.empty((num_dst_nodes, node_feat_dim + edge_feat_dim), device=offsets.device, dtype=offsets.dtype)
    else:
        out_positions = torch.empty((0,), device=offsets.device, dtype=offsets.dtype)

    if num_dst_nodes == 0:
        return output, out_positions

    _require_backend()
    import cudnn

    with torch.cuda.device(offsets.device):
        cudnn.gnn_agg_simple_forward(
            torch.cuda.current_stream(offsets.device).cuda_stream,
            offsets.data_ptr(),
            indices.data_ptr(),
            tensor_pointer(map_csc_to_coo),
            num_src_nodes,
            num_dst_nodes,
            num_edges,
            TORCH_INDEX_DTYPE_TO_CUDNN[offsets.dtype],
            tensor_pointer(node_features),
            tensor_pointer(edge_features),
            tensor_pointer(concat_features),
            output.data_ptr(),
            out_positions.data_ptr() if out_positions.numel() else 0,
            node_feat_dim,
            edge_feat_dim,
            concat_feat_dim,
            TORCH_DTYPE_TO_CUDNN[dtype],
            _AGGREGATION_TO_INT[aggr],
        )
    return output, out_positions


def backward(
    grad_output: Tensor,
    offsets: Tensor,
    indices: Tensor,
    map_csc_to_coo: Optional[Tensor],
    out_positions: Tensor,
    num_src_nodes: int,
    node_feat_dim: int,
    edge_feat_dim: int,
    concat_feat_dim: int,
    aggr: str,
) -> Tuple[Tensor, Tensor, Tensor]:
    _validate_aggregation(aggr)
    num_dst_nodes, num_edges = _validate_graph(offsets, indices, map_csc_to_coo, num_src_nodes)
    expected_shape = (num_dst_nodes, node_feat_dim + edge_feat_dim + concat_feat_dim)
    if tuple(grad_output.shape) != expected_shape:
        raise ValueError(f"grad_output must have shape {expected_shape}, got {tuple(grad_output.shape)}")
    if grad_output.dtype not in TORCH_DTYPE_TO_CUDNN or not grad_output.is_cuda:
        raise TypeError(f"grad_output must be a CUDA float32, float16, or bfloat16 tensor, got {grad_output.dtype}")
    if grad_output.device != offsets.device:
        raise ValueError(f"grad_output must be on {offsets.device}, got {grad_output.device}")

    if aggr in ("max", "min"):
        expected_positions_shape = (num_dst_nodes, node_feat_dim + edge_feat_dim)
        if tuple(out_positions.shape) != expected_positions_shape or out_positions.dtype != offsets.dtype:
            raise ValueError(
                f"out_positions must have shape {expected_positions_shape} and dtype {offsets.dtype}, "
                f"got {tuple(out_positions.shape)} and {out_positions.dtype}"
            )

    offsets = offsets.contiguous()
    indices = indices.contiguous()
    map_csc_to_coo = None if map_csc_to_coo is None else map_csc_to_coo.contiguous()
    grad_output = grad_output.contiguous()
    out_positions = out_positions.contiguous()

    if num_dst_nodes == 0:
        return (
            torch.zeros((num_src_nodes, node_feat_dim), device=offsets.device, dtype=grad_output.dtype),
            torch.empty((num_edges, edge_feat_dim), device=offsets.device, dtype=grad_output.dtype),
            torch.empty((num_dst_nodes, concat_feat_dim), device=offsets.device, dtype=grad_output.dtype),
        )

    # TODO: Use torch.empty once cuDNN AggSimple backward initializes every
    # gradient output buffer before accumulating into it.
    grad_node = torch.zeros((num_src_nodes, node_feat_dim), device=offsets.device, dtype=grad_output.dtype)
    grad_edge = torch.zeros((num_edges, edge_feat_dim), device=offsets.device, dtype=grad_output.dtype)
    grad_concat = torch.zeros((num_dst_nodes, concat_feat_dim), device=offsets.device, dtype=grad_output.dtype)

    _require_backend()
    import cudnn

    with torch.cuda.device(offsets.device):
        cudnn.gnn_agg_simple_backward(
            torch.cuda.current_stream(offsets.device).cuda_stream,
            offsets.data_ptr(),
            indices.data_ptr(),
            tensor_pointer(map_csc_to_coo),
            num_src_nodes,
            num_dst_nodes,
            num_edges,
            TORCH_INDEX_DTYPE_TO_CUDNN[offsets.dtype],
            grad_output.data_ptr(),
            out_positions.data_ptr() if out_positions.numel() else 0,
            grad_node.data_ptr() if node_feat_dim else 0,
            grad_edge.data_ptr() if edge_feat_dim else 0,
            grad_concat.data_ptr() if concat_feat_dim else 0,
            node_feat_dim,
            edge_feat_dim,
            concat_feat_dim,
            TORCH_DTYPE_TO_CUDNN[grad_output.dtype],
            _AGGREGATION_TO_INT[aggr],
        )
    return grad_node, grad_edge, grad_concat


_lib = torch.library.Library("cudnn", "FRAGMENT")
_lib.define(
    "gnn_agg_simple_fwd(Tensor offsets, Tensor indices, Tensor? map_csc_to_coo, Tensor? node_features, "
    "Tensor? edge_features, Tensor? concat_features, int num_src_nodes, str aggr) -> (Tensor, Tensor)"
)
_lib.define(
    "gnn_agg_simple_bwd(Tensor grad_output, Tensor offsets, Tensor indices, Tensor? map_csc_to_coo, Tensor out_positions, "
    "int num_src_nodes, int node_feat_dim, int edge_feat_dim, int concat_feat_dim, str aggr) -> (Tensor, Tensor, Tensor)"
)
_lib.impl("gnn_agg_simple_fwd", forward, "CUDA")
_lib.impl("gnn_agg_simple_bwd", backward, "CUDA")


@torch.library.register_fake("cudnn::gnn_agg_simple_fwd")
def _fwd_fake(
    offsets: Tensor,
    indices: Tensor,
    map_csc_to_coo: Optional[Tensor],
    node_features: Optional[Tensor],
    edge_features: Optional[Tensor],
    concat_features: Optional[Tensor],
    num_src_nodes: int,
    aggr: str,
) -> Tuple[Tensor, Tensor]:
    num_dst_nodes, _, dtype, node_feat_dim, edge_feat_dim, concat_feat_dim = _validate_inputs(
        offsets, indices, map_csc_to_coo, node_features, edge_features, concat_features, num_src_nodes, aggr
    )
    output = torch.empty((num_dst_nodes, node_feat_dim + edge_feat_dim + concat_feat_dim), device=offsets.device, dtype=dtype)
    positions_shape = (num_dst_nodes, node_feat_dim + edge_feat_dim) if aggr in ("max", "min") else (0,)
    return output, torch.empty(positions_shape, device=offsets.device, dtype=offsets.dtype)


@torch.library.register_fake("cudnn::gnn_agg_simple_bwd")
def _bwd_fake(
    grad_output: Tensor,
    offsets: Tensor,
    indices: Tensor,
    map_csc_to_coo: Optional[Tensor],
    out_positions: Tensor,
    num_src_nodes: int,
    node_feat_dim: int,
    edge_feat_dim: int,
    concat_feat_dim: int,
    aggr: str,
) -> Tuple[Tensor, Tensor, Tensor]:
    del map_csc_to_coo, out_positions, aggr
    num_dst_nodes = offsets.numel() - 1
    num_edges = indices.numel()
    return (
        torch.empty((num_src_nodes, node_feat_dim), device=grad_output.device, dtype=grad_output.dtype),
        torch.empty((num_edges, edge_feat_dim), device=grad_output.device, dtype=grad_output.dtype),
        torch.empty((num_dst_nodes, concat_feat_dim), device=grad_output.device, dtype=grad_output.dtype),
    )


def _setup_context(ctx, inputs, output) -> None:
    offsets, indices, map_csc_to_coo, node_features, edge_features, concat_features, num_src_nodes, aggr = inputs
    _, out_positions = output
    ctx.save_for_backward(offsets, indices, map_csc_to_coo, out_positions)
    ctx.mark_non_differentiable(out_positions)
    ctx.num_src_nodes = num_src_nodes
    ctx.node_feat_dim = 0 if node_features is None else node_features.shape[1]
    ctx.edge_feat_dim = 0 if edge_features is None else edge_features.shape[1]
    ctx.concat_feat_dim = 0 if concat_features is None else concat_features.shape[1]
    ctx.has_node_features = node_features is not None
    ctx.has_edge_features = edge_features is not None
    ctx.has_concat_features = concat_features is not None
    ctx.aggr = aggr


@torch.compiler.allow_in_graph
def _autograd_backward(ctx, grad_output: Tensor, grad_out_positions: Optional[Tensor]):
    del grad_out_positions
    offsets, indices, map_csc_to_coo, out_positions = ctx.saved_tensors
    grad_node, grad_edge, grad_concat = torch.ops.cudnn.gnn_agg_simple_bwd(
        grad_output,
        offsets,
        indices,
        map_csc_to_coo,
        out_positions,
        ctx.num_src_nodes,
        ctx.node_feat_dim,
        ctx.edge_feat_dim,
        ctx.concat_feat_dim,
        ctx.aggr,
    )
    return (
        None,
        None,
        None,
        grad_node if ctx.has_node_features else None,
        grad_edge if ctx.has_edge_features else None,
        grad_concat if ctx.has_concat_features else None,
        None,
        None,
    )


torch.library.register_autograd("cudnn::gnn_agg_simple_fwd", _autograd_backward, setup_context=_setup_context)
# Public API


def agg_simple(
    graph: CscGraph,
    *,
    node_features: Optional[Tensor] = None,
    edge_features: Optional[Tensor] = None,
    concat_features: Optional[Tensor] = None,
    aggr: str = "sum",
) -> Tensor:
    """Aggregate CSC-neighbor node/edge features and optionally append destination features."""

    output, _ = torch.ops.cudnn.gnn_agg_simple_fwd(
        graph.offsets,
        graph.indices,
        graph.map_csc_to_coo,
        node_features,
        edge_features,
        concat_features,
        graph.num_src_nodes,
        aggr,
    )
    return output
